"""Provides 'Var' dataclass"""

from dataclasses import dataclass
from typing import Optional

from nawah.enums import VarType


@dataclass(kw_only=True)
class Var:
    """Var dataclass serves role of placeholder variable value. It is used to set values in Config
    Attrs dynamically"""

    # pylint: disable=invalid-name

    type: VarType
    var: Optional[str] = None

    @staticmethod
    def ENV(var: str, /) -> "Var":
        """Shorthand method to define 'Var' object of type 'ENV'"""
        return Var(type=VarType.ENV, var=var)

    @staticmethod
    def SESSION(var: str, /) -> "Var":
        """Shorthand method to define 'Var' object of type 'SESSION'"""
        return Var(type=VarType.SESSION, var=var)

    @staticmethod
    def DOC(var: str, /) -> "Var":
        """Shorthand method to define 'Var' object of type 'DOC'"""
        return Var(type=VarType.DOC, var=var)

    @staticmethod
    def L10N(var: str, /) -> "Var":
        """Shorthand method to define 'Var' object of type 'L10N'"""
        return Var(type=VarType.L10N, var=var)

    @staticmethod
    def CONFIG(var: str, /) -> "Var":
        """Shorthand method to define 'Var' object of type 'CONFIG'"""
        return Var(type=VarType.CONFIG, var=var)

    @staticmethod
    def DATE() -> "Var":
        """Shorthand method to define 'Var' object of type 'DATE'"""
        return Var(type=VarType.DATE)

    @staticmethod
    def TIME() -> "Var":
        """Shorthand method to define 'Var' object of type 'TIME'"""
        return Var(type=VarType.TIME)

    @staticmethod
    def DATETIME() -> "Var":
        """Shorthand method to define 'Var' object of type 'DATETIME'"""
        return Var(type=VarType.DATETIME)

    def __post_init__(self):
        if self.type in [
            VarType.ENV,
            VarType.SESSION,
            VarType.DOC,
            VarType.L10N,
            VarType.CONFIG,
        ]:
            if not self.var:
                raise Exception(
                    f"Var object of type '{self.type}' should have value for 'var' attr"
                )



"""Provides exceptions used in Nawah"""

from ._attr import (InvalidAttrTypeArgException, InvalidAttrTypeException,
                    InvalidAttrTypeRefException, JSONPathNotFoundException)
from ._call import (InvalidCallArgException, InvalidCallEndpointException,
                    InvalidFuncException, InvalidModuleException,
                    NotPermittedException, UnknownCallArgException)
from ._config import ConfigException
from ._data import UnknownDeleteStrategyException
from ._func import (FuncException, InvalidDocAttrException,
                    InvalidDocException, InvalidQueryAttrException,
                    InvalidQueryException, MissingDocAttrException,
                    MissingQueryAttrException)
from ._query import (InvalidQueryOperTypeException,
                     InvalidQueryStepAttrLenException,
                     InvalidQueryStepAttrTypeException,
                     InvalidQueryStepLenException,
                     InvalidQueryStepTypeException, UnknownQueryOperException)
from ._validate import InvalidAttrException, MissingAttrException
from ._var import InvalidLocaleException, InvalidVarException

__all__ = [
    "InvalidAttrTypeArgException",
    "InvalidAttrTypeException",
    "InvalidAttrTypeRefException",
    "JSONPathNotFoundException",
    "InvalidCallArgException",
    "InvalidCallEndpointException",
    "InvalidFuncException",
    "InvalidModuleException",
    "NotPermittedException",
    "UnknownCallEndpointException",
    "ConfigException",
    "UnknownDeleteStrategyException",
    "FuncException",
    "InvalidDocAttrException",
    "InvalidDocException",
    "InvalidQueryAttrException",
    "InvalidQueryException",
    "MissingDocAttrException",
    "MissingQueryAttrException",
    "InvalidQueryOperTypeException",
    "InvalidQueryStepAttrLenException",
    "InvalidQueryStepAttrTypeException",
    "InvalidQueryStepLenException",
    "InvalidQueryStepTypeException",
    "UnknownQueryOperException",
    "InvalidAttrException",
    "MissingAttrException",
    "InvalidLocaleException",
    "InvalidVarException",
]



import datetime
import json
import logging
import re
from typing import TYPE_CHECKING, Any

import aiohttp.web
from multidict import MultiDict
from passlib.hash import pbkdf2_sha512
from requests_toolbelt.multipart import decoder

from nawah import data as Data
from nawah.classes import Query, app_encoder
from nawah.config import Config
from nawah.enums import Event
from nawah.utils import camel_to_upper

from .._call import _convert_deprecated_query, call
from .._config import _compile_anon_session, _compile_anon_user

if TYPE_CHECKING:
    from nawah.types import NawahDoc, NawahSession

logger = logging.getLogger("nawah")


async def _http_handler(request: aiohttp.web.Request):
    headers = MultiDict(
        [
            ("Server", "Nawah"),
            ("Powered-By", "Nawah, https://nawah.masaar.com"),
            ("Access-Control-Allow-Origin", "*"),
            ("Access-Control-Allow-Methods", "GET,POST,OPTIONS"),
            (
                "Access-Control-Allow-Headers",
                "Content-Type,X-Auth-Bearer,X-Auth-Token,X-Auth-App",
            ),
            ("Access-Control-Expose-Headers", "Content-Disposition"),
        ]
    )

    logger.debug("Received new %s request: %s", request.method, request.match_info)

    if request.method == "OPTIONS":
        return aiohttp.web.Response(
            status=200,
            headers=headers,
            body=app_encoder.encode(
                {
                    "status": 200,
                    "msg": "OPTIONS request is allowed",
                }
            ),
        )

    # Check for IP quota
    if str(request.remote) not in Config.sys.ip_quota:
        Config.sys.ip_quota[str(request.remote)] = {
            "counter": Config.quota_ip_min,
            "last_check": datetime.datetime.utcnow(),
        }
    else:
        if (
            datetime.datetime.utcnow()
            - Config.sys.ip_quota[str(request.remote)]["last_check"]
        ).seconds > 259:
            Config.sys.ip_quota[str(request.remote)][
                "last_check"
            ] = datetime.datetime.utcnow()
            Config.sys.ip_quota[str(request.remote)]["counter"] = Config.quota_ip_min
        else:
            if Config.sys.ip_quota[str(request.remote)]["counter"] - 1 <= 0:
                logger.warning(
                    "Denying '%s' request from '%s' for hitting IP quota",
                    request.method,
                    request.remote,
                )
                headers["Content-Type"] = "application/json; charset=utf-8"
                return aiohttp.web.Response(
                    status=429,
                    headers=headers,
                    body=app_encoder.encode(
                        {
                            "status": 429,
                            "msg": "You have hit calls quota from this IP",
                            "args": {"code": "CORE_REQ_IP_QUOTA_HIT"},
                        }
                    ),
                )

            Config.sys.ip_quota[str(request.remote)]["counter"] -= 1

    module_name = request.url.parts[1]
    func_name = request.url.parts[2]
    request_args = dict(request.match_info.items())

    data = Data.create_conn()
    session: "NawahSession" = {
        "conn": {
            "data": data,
            "REMOTE_ADDR": request.remote or "localhost",
            "HTTP_USER_AGENT": "",
            "HTTP_ORIGIN": "",
            "client_app": "__public",
            "args": {},
        }
    }

    session["conn"]["HTTP_USER_AGENT"] = (
        request.headers["user-agent"] if "user-agent" in request.headers else ""
    )
    session["conn"]["HTTP_ORIGIN"] = (
        request.headers["origin"] if "origin" in request.headers else ""
    )

    # [TODO] Add condition to now allow a connection from a client with client_app = __sys
    if Config.client_apps:
        if request.headers["X-Auth-App"] not in Config.client_apps or (
            Config.client_apps[request.headers["X-Auth-App"]].type == "web"
            and session["conn"]["HTTP_ORIGIN"]
            not in Config.client_apps[request.headers["X-Auth-App"]].origin
        ):
            logger.debug("Denying request due to unauthorised client_app")
            headers["Content-Type"] = "application/json; charset=utf-8"
            return aiohttp.web.Response(
                status=403,
                headers=headers,
                body=app_encoder.encode(
                    {
                        "status": 403,
                        "msg": "X-Auth headers could not be verified",
                        "args": {"code": "CORE_SESSION_INVALID_XAUTH"},
                    }
                ).encode("utf-8"),
            )

        session["conn"]["client_app"] = request.headers["X-Auth-App"]

    if " X-Auth-Bearer" in request.headers or "X-Auth-Token" in request.headers:
        logger.debug("Detected 'X-Auth' header[s]")
        if (
            "X-Auth-Bearer" not in request.headers
            or "X-Auth-Token" not in request.headers
            or "X-Auth-App" not in request.headers
        ):
            logger.debug("Denying request due to missing 'X-Auth' header")
            headers["Content-Type"] = "application/json; charset=utf-8"
            return aiohttp.web.Response(
                status=400,
                headers=headers,
                body=app_encoder.encode(
                    {
                        "status": 400,
                        "msg": "One 'X-Auth' headers was set but not the other",
                    }
                ).encode("utf-8"),
            )
        try:
            session_results = await call(
                "session/read",
                skip_events=[Event.PERM],
                session=session,
                query=Query(
                    [
                        {
                            "_id": {"$eq": request.headers["X-Auth-Bearer"]},
                        }
                    ]
                ),
            )
        except Exception as e:  # pylint: disable=broad-except
            headers["Content-Type"] = "application/json; charset=utf-8"
            if Config.debug:
                return aiohttp.web.Response(
                    status=500,
                    headers=headers,
                    body=app_encoder.encode(
                        {
                            "status": 500,
                            "msg": f"Unexpected error has occurred [{e}]",
                            "args": {
                                "code": "CORE_SERVER_ERROR",
                                "err": str(e),
                            },
                        }
                    ).encode("utf-8"),
                )

            return aiohttp.web.Response(
                status=500,
                headers=headers,
                body=app_encoder.encode(
                    {
                        "status": 500,
                        "msg": "Unexpected error has occurred",
                        "args": {"code": "CORE_SERVER_ERROR"},
                    }
                ).encode("utf-8"),
            )

        if not session_results["args"]["count"] or not pbkdf2_sha512.verify(
            request.headers["X-Auth-Token"],
            session_results["args"]["docs"][0]["token_hash"],
        ):
            logger.debug("Denying request due to missing failed Call Authorisation")
            headers["Content-Type"] = "application/json; charset=utf-8"
            return aiohttp.web.Response(
                status=403,
                headers=headers,
                body=app_encoder.encode(
                    {
                        "status": 403,
                        "msg": "X-Auth headers could not be verified",
                        "args": {"code": "CORE_SESSION_INVALID_XAUTH"},
                    }
                ).encode("utf-8"),
            )
        user_session = session_results["args"]["docs"][0]
        session_results = await call(
            "session/reauth",
            skip_events=[Event.PERM],
            session=session,
            query=Query(
                [
                    {"_id": {"$eq": request.headers["X-Auth-Bearer"]}},
                    {"token": {"$eq": request.headers["X-Auth-Token"]}},
                    {"groups": {"$eq": []}},
                ]
            ),
        )
        if session_results["status"] != 200:
            logger.debug("Denying request due to fail to reauth")
            headers["Content-Type"] = "application/json; charset=utf-8"
            return aiohttp.web.Response(
                status=403,
                headers=headers,
                body=app_encoder.encode(session_results).encode("utf-8"),
            )

        user_session = session_results["args"]["session"]
    else:
        anon_user = _compile_anon_user()
        anon_session = _compile_anon_session()
        anon_session["user"] = anon_user
        user_session = anon_session

    session.update(json.loads(app_encoder.encode(user_session)))

    doc: "NawahDoc" = {}

    if "Content-Type" in request.headers:
        doc_content = await request.content.read()
        # Check Content-Type to decide how to deserialise content
        if request.headers["Content-Type"].startswith("multipart/form-data"):
            multipart_content_type = request.headers["Content-Type"]
            doc = {}
            try:
                multipart_parts = decoder.MultipartDecoder(
                    doc_content, multipart_content_type
                ).parts

                for part in multipart_parts:
                    content_disposition = part.headers[b"Content-Disposition"].decode(
                        "UTF-8"
                    )
                    attr_name_match = re.findall('name="([^"]+)"', content_disposition)
                    if not attr_name_match:
                        continue
                    attr_name = attr_name_match[0]
                    doc[attr_name] = part.content
                    # For non-file values, decode unto str
                    if "filename" not in content_disposition:
                        doc[attr_name] = doc[attr_name].decode("UTF-8")
            except decoder.ImproperBodyPartContentException:
                logger.debug(
                    "Ignoring processing 'doc_content' due to 'decoder."
                    "ImproperBodyPartContentException' Exception"
                )

        elif request.headers["Content-Type"].startswith("application/json"):
            doc = json.loads(doc_content)

    # Attempt to generate value for query
    query = None
    # If request_args (from GET route path) has length, use it as query
    if len(request_args):
        query = Query([{k: {"$eq": v}} for k, v in request_args.items()])  # type: ignore
    # Otherwise, check if doc contains $query
    else:
        if "$query" in doc:
            logger.debug(
                "Detected '$query' attr in doc. Attempting to process it as Query"
            )
            try:
                if isinstance(doc["$query"], list):
                    logger.warning("Call is using deprecated query format: %s", query)
                    query_special = {}
                    for step in doc["$query"]:
                        if not isinstance(step, dict):
                            continue
                        for attr_name, attr_val in step.items():
                            if isinstance(attr_name, str) and attr_name[0] == "$":
                                query_special[attr_name] = attr_val
                    query = Query(
                        _convert_deprecated_query(doc["$query"]), special=query_special
                    )
                else:
                    special: Any = {}
                    if "$special" in doc["$query"]:
                        special = doc["$query"]["$special"]
                    if "$pipe" in doc["$query"]:
                        query = Query(doc["$query"]["$pipe"], special=special)
                    else:
                        query = Query(special=special)
            except Exception as e:  # pylint: disable=broad-except
                logger.error("Failed to process '$query' doc attr. Exception: %s", e)

            del doc["$query"]

    try:
        results = await call(
            f"{module_name}/{func_name}",
            session=session,
            query=query,
            doc=doc,
        )
    except Exception as e:  # pylint: disable=broad-except
        status = getattr(e, "status", 500)
        msg = "Unexpected error has occurred"
        code = camel_to_upper(e.__class__.__name__).replace("_EXCEPTION", "")
        if status != 500 and e.args:
            msg = e.args[0]

        results = {
            "status": status,
            "msg": msg,
            "args": {"code": code},
        }

    logger.debug("Closing connection")
    session["conn"]["data"].close()

    if "return" not in results["args"] or results["args"]["return"] == "json":
        if "return" in results["args"]:
            del results["args"]["return"]
        headers["Content-Type"] = "application/json; charset=utf-8"
        if results["status"] == 404:
            return aiohttp.web.Response(
                status=results["status"],
                headers=headers,
                body=app_encoder.encode(
                    {"status": 404, "msg": "Requested content not found"}
                ).encode("utf-8"),
            )
        return aiohttp.web.Response(
            status=results["status"],
            headers=headers,
            body=app_encoder.encode(results),
        )

    if results["args"]["return"] == "file":
        del results["args"]["return"]
        status = results["status"]
        expiry_time = datetime.datetime.utcnow() + datetime.timedelta(days=30)
        headers["Last-Modified"] = str(results["args"]["docs"][0]["lastModified"])
        headers["Content-Type"] = results["args"]["docs"][0]["type"]
        headers["Cache-Control"] = "public, max-age=31536000"
        headers["Expires"] = expiry_time.strftime("%a, %d %b %Y %H:%M:%S GMT")
        content = results["args"]["docs"][0]["content"]
        if "If-Modified-Since" in request.headers:
            if request.headers["If-Modified-Since"] == headers["Last-Modified"]:
                status = 304
                content = b""
        return aiohttp.web.Response(
            status=status,
            headers=headers,
            body=content,
        )

    if results["args"]["return"] == "msg":
        del results["args"]["return"]
        headers["Content-Type"] = "application/json; charset=utf-8"
        return aiohttp.web.Response(
            status=results["status"], headers=headers, body=results["msg"]
        )

    headers["Content-Type"] = "application/json; charset=utf-8"
    return aiohttp.web.Response(
        status=405,
        headers=headers,
        body=app_encoder.encode({"status": 405, "msg": "405 NOT ALLOWED"}),
    )



from typing import TYPE_CHECKING, Literal

from nawah.config import Config

if TYPE_CHECKING:
    from nawah.types import NawahSession


async def drop(session: "NawahSession", collection_name: str) -> Literal[True]:
    collection = session["conn"]["data"][Config.data_name][collection_name]
    await collection.drop()
    return True



"""Provides Nawah Utilities which provide functionality for internal and external behaviours"""

from ._attr_type import (
    decode_attr_type,
    encode_attr_type,
    extract_attr,
    generate_attr_val,
)
from ._cache import update_cache
from ._call import call
from ._val import (
    camel_to_upper,
    deep_update,
    expand_val,
    extract_val,
    set_val,
    var_value,
)
from ._validate import validate_attr, validate_doc, validate_type

__all__ = [
    "decode_attr_type",
    "encode_attr_type",
    "extract_attr",
    "generate_attr_val",
    "update_cache",
    "call",
    "camel_to_upper",
    "deep_update",
    "expand_val",
    "extract_val",
    "set_val",
    "var_value",
    "validate_attr",
    "validate_doc",
    "validate_type",
]



"""Provides types used in Nawah"""

from ._types import (AnalyticsEvents, AppPackage, IPQuota, NawahConn,
                     NawahConnQuota, NawahDoc, NawahEvents, NawahQueryOperAnd,
                     NawahQueryOperOr, NawahQuerySpecial,
                     NawahQuerySpecialGroup, NawahQueryStep, NawahSession,
                     Results, ResultsArgs)

__all__ = [
    "AnalyticsEvents",
    "AppPackage",
    "IPQuota",
    "NawahConn",
    "NawahConnQuota",
    "NawahDoc",
    "NawahEvents",
    "NawahQueryOperAnd",
    "NawahQueryOperOr",
    "NawahQuerySpecial",
    "NawahQuerySpecialGroup",
    "NawahQueryStep",
    "NawahSession",
    "Results",
    "ResultsArgs",
]



import asyncio
import datetime
import json
import logging
import random
import traceback
from functools import partial
from typing import TYPE_CHECKING, cast

import aiohttp.web

from nawah import data as Data
from nawah.classes import app_encoder
from nawah.config import Config

from .._call import call
from .._config import _compile_anon_session, _compile_anon_user
from .._val import camel_to_upper
from ._utils import _close_session

if TYPE_CHECKING:
    from asyncio.futures import Future

    from nawah.types import NawahSession

logger = logging.getLogger("nawah")


async def _websocket_handler(request: aiohttp.web.Request):
    data = Data.create_conn()
    logger.debug("Websocket connection starting with client at '%s'", request.remote)
    ws = aiohttp.web.WebSocketResponse()
    await ws.prepare(request)

    session_lock_key = random.random()

    while Config.sys.session_lock != session_lock_key:
        if Config.sys.session_lock is None:
            Config.sys.session_lock = session_lock_key

    Config.sys.session_counter += 1

    session = _compile_anon_session()
    session["user"] = _compile_anon_user()
    session["conn"] = {
        "id": str(Config.sys.session_counter),
        "data": data,
        "REMOTE_ADDR": request.remote or "localhost",
        "HTTP_USER_AGENT": "",
        "HTTP_ORIGIN": "",
        "client_app": "__public",
        "ws": ws,
        "init": False,
        "last_call": datetime.datetime.utcnow(),
        "quota": {
            "counter": Config.quota_anon_min,
            "last_check": datetime.datetime.utcnow(),
        },
        "args": {},
    }

    Config.sys.sessions[str(Config.sys.session_counter)] = session
    Config.sys.session_lock = None

    session["conn"]["HTTP_USER_AGENT"] = (
        request.headers["user-agent"] if "user-agent" in request.headers else ""
    )
    session["conn"]["HTTP_ORIGIN"] = (
        request.headers["origin"] if "origin" in request.headers else ""
    )

    logger.debug(
        "Websocket connection #'%s' ready with client at '%s'",
        session["conn"]["id"],
        session["conn"]["REMOTE_ADDR"],
    )

    await ws.send_str(
        app_encoder.encode(
            {
                "status": 200,
                "msg": "Connection ready",
                "args": {"code": "CORE_CONN_READY"},
            }
        )
    )

    async for msg in ws:
        if "data" not in session["conn"]:
            await ws.close()
            break
        logger.debug(
            "Received new message from session #'%s': %s",
            session["conn"]["id"],
            msg.data[:256],
        )
        if msg.type == aiohttp.WSMsgType.TEXT:
            logger.debug(
                "Config.sys.ip_quota on session #'%s': %s",
                session["conn"]["id"],
                Config.sys.ip_quota,
            )
            logger.debug(
                "session_quota: on session #'%s': %s",
                session["conn"]["id"],
                session["conn"]["quota"],
            )
            # Check for IP quota
            if str(request.remote) not in Config.sys.ip_quota:
                Config.sys.ip_quota[str(request.remote)] = {
                    "counter": Config.quota_ip_min,
                    "last_check": datetime.datetime.utcnow(),
                }
            else:
                if (
                    datetime.datetime.utcnow()
                    - Config.sys.ip_quota[str(request.remote)]["last_check"]
                ).seconds > 59:
                    Config.sys.ip_quota[str(request.remote)][
                        "last_check"
                    ] = datetime.datetime.utcnow()
                    Config.sys.ip_quota[str(request.remote)][
                        "counter"
                    ] = Config.quota_ip_min
                else:
                    # If hit quota, deny call
                    if Config.sys.ip_quota[str(request.remote)]["counter"] - 1 <= 0:
                        logger.warning(
                            "Denying Websocket request from '%s' for hitting IP quota",
                            request.remote,
                        )
                        asyncio.create_task(
                            _handle_msg(
                                session=session,
                                msg=msg,
                                decline_quota="ip",
                            )
                        )
                        continue
                    # Else, update quota counter
                    Config.sys.ip_quota[str(request.remote)]["counter"] -= 1
            # Check for session quota
            if (
                datetime.datetime.utcnow() - session["conn"]["quota"]["last_check"]
            ).seconds > 59:
                session["conn"]["quota"]["last_check"] = datetime.datetime.utcnow()
                session["conn"]["quota"]["counter"] = (
                    (Config.quota_anon_min - 1)
                    if not session or session["token"] == Config.anon_token
                    else (Config.quota_auth_min - 1)
                )
                asyncio.create_task(_handle_msg(session=session, msg=msg))
            else:
                # If hit quota, deny call
                if session["conn"]["quota"]["counter"] - 1 <= 0:
                    asyncio.create_task(
                        _handle_msg(
                            session=session,
                            msg=msg,
                            decline_quota="session",
                        )
                    )
                    continue
                # Else, update quota counter
                session["conn"]["quota"]["counter"] -= 1
                asyncio.create_task(_handle_msg(session=session, msg=msg))

    if "conn" in session["conn"]:
        await _close_session(session["conn"]["id"])

    return ws


async def _handle_msg(
    session: "NawahSession",
    msg: aiohttp.WSMessage,
    decline_quota: str = None,
):
    try:
        session["conn"]["last_call"] = datetime.datetime.utcnow()
        res = json.loads(msg.data)

        # Check if msg should be denied for quota hit
        if decline_quota == "ip":
            await session["conn"]["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": 429,
                        "msg": "You have hit calls quota from this IP",
                        "args": {
                            "call_id": res["call_id"]
                            if "call_id" in res.keys()
                            else None,
                            "code": "CORE_REQ_IP_QUOTA_HIT",
                        },
                    }
                )
            )
            return

        if decline_quota == "session":
            await session["conn"]["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": 429,
                        "msg": "You have hit calls quota",
                        "args": {
                            "call_id": res["call_id"]
                            if "call_id" in res.keys()
                            else None,
                            "code": "CORE_REQ_SESSION_QUOTA_HIT",
                        },
                    }
                )
            )
            return

        logger.debug("Decoded request: %s", app_encoder.encode(res))

        if "endpoint" not in res.keys():
            await session["conn"]["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": 400,
                        "msg": "Request missing endpoint",
                        "args": {
                            "call_id": res["call_id"]
                            if "call_id" in res.keys()
                            else None,
                            "code": "CORE_REQ_NO_ENDPOINT",
                        },
                    }
                )
            )
            return

        if session["conn"]["init"] is False:
            if res["endpoint"] != "conn/verify":
                await session["conn"]["ws"].send_str(
                    app_encoder.encode(
                        {
                            "status": 1008,
                            "msg": "Request token is not accepted",
                            "args": {
                                "call_id": res["call_id"]
                                if "call_id" in res.keys()
                                else None,
                                "code": "CORE_REQ_NO_VERIFY",
                            },
                        }
                    )
                )
                await session["conn"]["ws"].close()
                return

            # Test if Nawah App defines list of approved client_apps
            if Config.client_apps:
                # Create testing sequence based on conn_verified boolean
                conn_verified = True
                # Test request includes client app name
                conn_verified = "doc" in res and "app" in res["doc"]
                # Continue testing if possible, and test ClientApp definition
                conn_verified = (
                    conn_verified
                    and (client_app_name := res["doc"]["app"]) in Config.client_apps
                )
                # Continue testing if possible, and test platform specific attributes
                if conn_verified:
                    if (
                        client_app := Config.client_apps[client_app_name]
                    ).type == "web":
                        if session["conn"]["HTTP_ORIGIN"] not in cast(
                            list[str], client_app.origin
                        ):
                            conn_verified = False
                # If not verified close connection
                if not conn_verified:
                    await session["conn"]["ws"].send_str(
                        app_encoder.encode(
                            {
                                "status": 1008,
                                "msg": "Request token is not accepted",
                                "args": {
                                    "call_id": res["call_id"]
                                    if "call_id" in res.keys()
                                    else None,
                                    "code": "CORE_REQ_NO_VERIFY",
                                },
                            }
                        )
                    )
                    await session["conn"]["ws"].close()
                    return

            session["conn"]["init"] = True
            if Config.client_apps:
                session["conn"]["client_app"] = res["doc"]["app"]
            logger.debug(
                "Connection on session #'%s' is verified", session["conn"]["id"]
            )
            # if Config.analytics_events['app_conn_verified']:
            #     asyncio.create_task(
            #         call(
            #             'analytic/create',
            #             skip_events=[Event.PERM],
            #             session=session,
            #             doc={
            #                 'event': 'CONN_VERIFIED',
            #                 'subevent': env['client_app'],
            #                 'args': {
            #                     'REMOTE_ADDR': env['REMOTE_ADDR'],
            #                     'HTTP_USER_AGENT': env['HTTP_USER_AGENT'],
            #                 },
            #             },
            #         )
            #     )
            await session["conn"]["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": 200,
                        "msg": "Connection established",
                        "args": {
                            "call_id": res["call_id"]
                            if "call_id" in res.keys()
                            else None,
                            "code": "CORE_CONN_OK",
                        },
                    }
                )
            )
            return

        if res["endpoint"] == "conn/close":
            logger.debug(
                "Received connection close instructions on session #'%s'",
                session["conn"]["id"],
            )
            await session["conn"]["ws"].close()
            return

        if res["endpoint"] == "heart/beat":
            logger.debug(
                "Received connection heartbeat on session #'%s'", session["conn"]["id"]
            )
            await session["conn"]["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": 200,
                        "msg": "Heartbeat received",
                        "args": {
                            "call_id": res["call_id"]
                            if "call_id" in res.keys()
                            else None,
                            "code": "CORE_HEARTBEAT_OK",
                        },
                    }
                )
            )
            return

        res["endpoint"] = res["endpoint"]

        if (
            res["endpoint"] in ["session/auth", "session/reauth"]
            and str(session["_id"]) != "f00000000000000000000012"
        ):
            await session["conn"]["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": 400,
                        "msg": "You are already authed",
                        "args": {
                            "call_id": res["call_id"]
                            if "call_id" in res.keys()
                            else None,
                            "code": "CORE_SESSION_ALREADY_AUTHED",
                        },
                    }
                )
            )
            return

        if (
            res["endpoint"] == "session/signout"
            and str(session["_id"]) == "f00000000000000000000012"
        ):
            await session["conn"]["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": 400,
                        "msg": "Singout is not allowed for '__ANON' user",
                        "args": {
                            "call_id": res["call_id"]
                            if "call_id" in res.keys()
                            else None,
                            "code": "CORE_SESSION_ANON_SIGNOUT",
                        },
                    }
                )
            )
            return

        if "query" not in res.keys():
            res["query"] = {}
        if "doc" not in res.keys():
            res["doc"] = {}
        if "call_id" not in res.keys():
            res["call_id"] = ""

        request = {
            "call_id": res["call_id"],
            # 'sid': res['sid'] or False,
            "query": res["query"],
            "doc": res["doc"],
            "path": res["endpoint"].split("/"),
        }

        if len(request["path"]) != 2:
            await session["conn"]["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": 400,
                        "msg": "Endpoint path is invalid",
                        "args": {
                            "call_id": request["call_id"],
                            "code": "CORE_REQ_INVALID_PATH",
                        },
                    }
                )
            )
            return

        module = request["path"][0]

        if module not in Config.modules:
            await session["conn"]["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": 400,
                        "msg": "Endpoint module is invalid",
                        "args": {
                            "call_id": request["call_id"],
                            "code": "CORE_REQ_INVALID_MODULE",
                        },
                    }
                )
            )
            return

        if request["path"][1] not in Config.modules[module].funcs:
            await session["conn"]["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": 400,
                        "msg": "Endpoint is invalid",
                        "args": {
                            "call_id": request["call_id"],
                            "code": "CORE_REQ_INVALID_METHOD",
                        },
                    }
                )
            )
            return

        if Config.modules[module].funcs[request["path"][1]].get_func:
            await session["conn"]["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": 400,
                        "msg": "Endpoint is a GET Function",
                        "args": {
                            "call_id": request["call_id"],
                            "code": "CORE_REQ_GET_FUNC",
                        },
                    }
                )
            )
            return

        # if not request['sid']:
        request["sid"] = "f00000000000000000000012"

        query = request["query"]
        doc = request["doc"]
        call_task = asyncio.create_task(
            call(
                f'{module}/{request["path"][1]}',
                skip_events=[],
                session=session,
                query=query,
                doc=doc,
            )
        )
        call_task.add_done_callback(
            partial(_call_callback, session=session, call_id=request["call_id"])
        )

    except Exception as e:
        logger.error("An error occurred. Details: %s", traceback.format_exc())
        if Config.debug:
            await session["conn"]["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": 500,
                        "msg": f"Unexpected error has occurred [{str(e)}]",
                        "args": {
                            "code": "CORE_SERVER_ERROR",
                            "err": str(e),
                            "call_id": request["call_id"],
                        },
                    }
                )
            )
        else:
            await session["conn"]["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": 500,
                        "msg": "Unexpected error has occurred",
                        "args": {
                            "code": "CORE_SERVER_ERROR",
                            "call_id": request["call_id"],
                        },
                    }
                )
            )


def _call_callback(call_task: "Future", session: "NawahSession", call_id: str):
    try:
        results = call_task.result()
        results["args"]["call_id"] = call_id
        asyncio.create_task(session["conn"]["ws"].send_str(app_encoder.encode(results)))
    except Exception as e:  # pylint: disable=broad-except
        status = getattr(e, "status", 500)
        msg = "Unexpected error has occurred"
        code = camel_to_upper(e.__class__.__name__).replace("_EXCEPTION", "")
        if status != 500 and e.args:
            msg = e.args[0]

        asyncio.create_task(
            session["conn"]["ws"].send_str(
                app_encoder.encode(
                    {
                        "status": status,
                        "msg": msg,
                        "args": {"code": code, "call_id": call_id},
                    }
                )
            )
        )



"""Provieds Utilities to be used to get and set cache"""

import copy
import datetime
import inspect
import logging
from typing import TYPE_CHECKING, Any, Optional, Protocol, Tuple, cast

import jwt
import redis.exceptions

from nawah.classes import app_encoder
from nawah.config import Config

if TYPE_CHECKING:
    from bson import ObjectId

    from nawah.classes import Cache, Func, Module, Query
    from nawah.types import NawahDoc, NawahEvents, NawahSession, Results

logger = logging.getLogger("nawah")

Config.cache_expiry = cast(int, Config.cache_expiry)


class CacheNotConfiguredException(Exception):
    """raises if a Cache Utility is called while app is not configured for Cache Workflow"""


class UpdateCacheRemoveCondition(Protocol):
    """Provides type-hint for \'update_cache\' Utility \'remove_condition\' callable"""

    # pylint: disable=too-few-public-methods

    def __call__(self, *, update_doc: "NawahDoc") -> bool:
        ...


async def check_cache_connection(attempt: int = 3):
    """Attempts to read from cache to force re-connection if broken"""

    if not Config.sys.cache:
        raise CacheNotConfiguredException()

    try:
        await Config.sys.cache.get("__connection")
    except redis.exceptions.ConnectionError as e:
        if attempt != 0:
            return await check_cache_connection(attempt=attempt - 1)

        raise e


async def reset_cache_channel(channel: str, /):
    """Resets specific cache `channel` by deleting it from active Redis db"""

    if not (cache := Config.sys.cache):
        raise CacheNotConfiguredException()

    await check_cache_connection()

    try:
        for key in await cache.client.keys(f"{channel}:*"):
            try:
                await cache.client.delete(key.decode("utf-8"))
            except redis.exceptions.ResponseError:
                logger.error("Failed to delete Cache Key: '%s'", key)
    except redis.exceptions.ConnectionError:
        logger.error(
            "Connection with Redis server '%s' failed. Skipping resetting Cache Channel '%s'.",
            Config.cache_server,
            channel,
        )


async def update_cache(
    *,
    channels: list[str],
    docs: list["ObjectId"],
    update_doc: "NawahDoc",
    remove_condition: "UpdateCacheRemoveCondition" = None,
):
    if not (cache := Config.sys.cache):
        raise CacheNotConfiguredException()

    remove_key = False
    try:
        if remove_condition:
            remove_key = remove_condition(update_doc=update_doc)
    except:
        remove_key = True

    for channel in channels:

        if channel.endswith(":"):
            channel += "*"
        elif not channel.endswith(":*"):
            channel += ":*"

        try:
            await check_cache_connection()

            for key in await cache.client.keys(channel):
                key = key.decode("utf-8")

                key_docs = await cache.get(key, ".docs")
                for doc in docs:
                    try:
                        doc_index = key_docs.index(str(doc))
                    except ValueError:
                        continue
                    if remove_key:
                        try:
                            await cache.delete(key, ".")
                        except redis.exceptions.ResponseError:
                            logger.error(
                                "Cache command failed with 'ResponseError'. Current scope:"
                            )
                            logger.error(locals())
                        continue
                    if await cache.get(
                        key, f".results.args.docs[{doc_index}]._id"
                    ) != str(doc):
                        try:
                            await cache.delete(key, ".")
                        except redis.exceptions.ResponseError as e:
                            logger.error(
                                "Cache command failed with 'ResponseError': '%s'",
                                e,
                            )
                            logger.error("Current scope: %s", locals())
                    try:
                        for attr_name in update_doc:
                            if update_doc[attr_name] is None:
                                continue
                            await cache.set(
                                key,
                                f".results.args.docs[{doc_index}].{attr_name}",
                                update_doc[attr_name],
                            )
                    except redis.exceptions.ResponseError as e:
                        logger.error(
                            "Cache command failed with 'ResponseError': '%s'",
                            e,
                        )
                        logger.error("Current scope: %s", locals())
                        logger.error(
                            "Removing key '%s' due to failed update attempt.", key
                        )
                        try:
                            await cache.delete(key, ".")
                        except redis.exceptions.ResponseError as ee:
                            logger.error(
                                "Cache command failed with 'ResponseError': '%s'",
                                ee,
                            )
                            logger.error("Current scope: %s", locals())
                try:
                    await cache.set(
                        key,
                        ".results.args.cache_time",
                        datetime.datetime.utcnow().isoformat(),
                    )
                except redis.exceptions.ResponseError as e:
                    logger.error(
                        "Cache command failed with 'ResponseError': '%s'",
                        e,
                    )
                    logger.error("Current scope: %s", locals())
        except redis.exceptions.ConnectionError:
            logger.error(
                "Connection with Redis server '%s' failed. Skipping updating Cache Channel '%s'.",
                Config.cache_server,
                channel,
            )


def _generate_cache_key(
    *,
    func: "Func",
    skip_events: "NawahEvents",
    session: "NawahSession",
    query: "Query",
) -> Optional[str]:
    if not Config.sys.cache or not func.cache:
        return None

    condition_params = {
        "skip_events": skip_events,
        "session": session,
        "query": query,
    }

    if not func.cache.condition(
        **{
            param: condition_params[param]
            for param in inspect.signature(func.cache.condition).parameters
        }
    ):
        return None

    cache_key = {
        "query": app_encoder.encode(query.pipe),
        "special": app_encoder.encode(query.special),
        "user": session["user"]["_id"] if func.cache.user_scoped else None,
    }

    cache_key_jwt = jwt.encode(cache_key, "_").split(".")[1]

    return cache_key_jwt


async def _call_cache(func: "Func", cache_key: str):
    if not Config.sys.cache:
        return

    module = cast("Module", func.module)
    cache = cast("Cache", func.cache)

    try:
        logger.debug(
            "Attempting to get cache with 'key': '%s'.",
            f"{module.name}:{func.name}:{cache_key}",
        )

        await check_cache_connection()

        if cache.file:
            cache_dict = None
            file_cache = await Config.sys.cache.client.hgetall(
                f"__files__:{cache.channel}:{module.name}:{func.name}:{cache_key}",
            )
            if file_cache:
                cache_dict = {
                    "status": 200,
                    "msg": "",
                    "args": {
                        "return": "file",
                        "docs": [
                            {
                                file_cache_key.decode("utf-8"): file_cache_val
                                if file_cache_key == b"content"
                                else file_cache_val.decode("utf-8")
                                for file_cache_key, file_cache_val in file_cache.items()
                            }
                        ],
                    },
                }

        else:
            cache_dict = await Config.sys.cache.get(
                f"{cache.channel}:{module.name}:{func.name}:{cache_key}",
                ".results",
            )

        logger.debug(
            "- Done getting cache with 'key': '%s'.",
            f"{module.name}:{func.name}:{cache_key}",
        )

        return cache_dict

    except redis.exceptions.ResponseError as e:
        logger.error(
            "Request to Redis server '%s' failed with ResponseError. Skipping Cache Workflow.",
            Config.cache_server,
        )
        logger.error("Redis error details: %s", e)
        return

    except redis.exceptions.ConnectionError:
        logger.error(
            "Connection with Redis server '%s' failed. Skipping Cache Workflow.",
            Config.cache_server,
        )
        return


async def _set_cache(*, func: "Func", cache_key: str, results: "Results"):
    if not Config.sys.cache:
        return

    module = cast("Module", func.module)
    cache = cast("Cache", func.cache)

    cache_key_long = f"{cache.channel}:{module.name}:{func.name}:{cache_key}"

    try:
        logger.debug(
            "Attempting to set cache with 'key': '%s'.",
            cache_key_long,
        )

        await check_cache_connection()

        cache_dict = {
            "docs": [doc["_id"] for doc in results["args"]["docs"]]
            if "args" in results and "docs" in results["args"]
            else [],
            "results": copy.deepcopy(results),
        }

        # Check if results contain file, and set using regular Redis cache
        if cache.file:
            # Update value for cache_key_long to prevent conflict with RedisJSON cache
            cache_key_long = f"__files__:{cache_key_long}"
            # Set cache
            await Config.sys.cache.client.hset(
                cache_key_long,
                mapping=results["args"]["docs"][0],
            )

        # Otherwise, set as RedisJSON cache
        else:
            await Config.sys.cache.set(
                cache_key_long,
                ".",
                cache_dict,
            )

        if Config.cache_expiry:
            await Config.sys.cache.client.expire(
                cache_key_long,
                Config.cache_expiry,
            )

        logger.debug(
            "- Done setting cache with 'key': '%s'.",
            cache_key_long,
        )

    except redis.exceptions.ConnectionError:
        logger.error(
            "Connection with Redis server '%s' failed. Skipping Cache Workflow.",
            Config.cache_server,
        )


async def _get_cache(
    *,
    func: "Func",
    skip_events: "NawahEvents",
    session: "NawahSession",
    query: "Query",
) -> Tuple[Optional[str], Any]:
    cache_key = _generate_cache_key(
        func=func,
        skip_events=skip_events,
        session=session,
        query=query,
    )
    call_cache = None

    if not cache_key:
        return (None, None)

    call_cache = await _call_cache(func=func, cache_key=cache_key)

    return (cache_key, call_cache)



"""Provides 'user' Module Functions callables"""

import copy
from typing import TYPE_CHECKING, Optional

from bson import ObjectId

from nawah.classes import Query
from nawah.config import Config
from nawah.enums import Event
from nawah.utils import call

from ._exceptions import (GroupAddedException, GroupNotAddedException,
                          InvalidGroupException, InvalidUserException)

if TYPE_CHECKING:
    from nawah.types import NawahDoc, NawahEvents, NawahSession, Results


async def _read(
    skip_events: "NawahEvents",
    session: "NawahSession",
    query: "Query",
    skip_sanitise_results: Optional[bool],
    raise_no_success: Optional[bool],
) -> "Results":
    skip_events.append(Event.PERM)
    read_results = await call(
        "base/read",
        module_name="user",
        skip_events=skip_events,
        session=session,
        query=query,
        args={
            "raise_no_success": raise_no_success,
        },
    )

    if read_results["status"] != 200:
        return read_results

    # Return results as is if both sanitise_results, user_settings are to be skipped
    if skip_sanitise_results is True:
        return read_results

    # Otherwise, iterate over results and apply not skipped action
    for i, _ in enumerate(read_results["args"]["docs"]):
        user = read_results["args"]["docs"][i]
        # Apply sanitise_results, removing hashes, if not skipped
        if skip_sanitise_results is not True:
            for user_attr_sanitise in Config.user_attrs_sanitise:
                del user[user_attr_sanitise]

    return read_results


async def _create(
    skip_events: "NawahEvents",
    session: "NawahSession",
    doc: "NawahDoc",
    raise_no_success: Optional[bool],
) -> "Results":
    if Event.ATTRS_DOC not in skip_events:
        doc["groups"] = [ObjectId("f00000000000000000000013")]

    create_results = await call(
        "base/create",
        skip_events=[Event.PERM],
        module_name="user",
        session=session,
        doc=doc,
        args={
            "raise_no_success": raise_no_success,
        },
    )

    return create_results


async def _read_privileges(session: "NawahSession", query: "Query") -> "Results":
    # Confirm _id is valid
    results = await call(
        "user/read",
        skip_events=[Event.PERM],
        session=session,
        query=Query([{"_id": {"$eq": query["_id:$eq"][0]}}]),
        args={"skip_sanitise_results": True},
    )

    if not results["args"]["count"]:
        raise InvalidUserException()

    user = results["args"]["docs"][0]
    # Loop over user's groups to extract privileges from
    for group in user["groups"]:
        group_results = await call(
            "group/read",
            skip_events=[Event.PERM],
            session=session,
            query=Query([{"_id": {"$eq": group}}]),
            args={
                "raise_no_success": True,
            },
        )
        group = group_results["args"]["docs"][0]
        for privilege in group["privileges"].keys():
            if privilege not in user["privileges"].keys():
                user["privileges"][privilege] = []
            for i in range(len(group["privileges"][privilege])):
                if (
                    group["privileges"][privilege][i]
                    not in user["privileges"][privilege]
                ):
                    user["privileges"][privilege].append(
                        group["privileges"][privilege][i]
                    )

    # Check for key asterisk in user's privileges
    if '*' in user["privileges"]:
        # Key asterisk exists, insert keys for all modules, not already added, and add key asterisk
        # value to it
        for module_name in Config.modules:
            if module_name not in user['privileges']:
                # For non-existing modules in user's privileges, copy key asterisk value
                user['privileges'][module_name] = copy.deepcopy(user['privileges']['*'])
            else:
                # Otherwise, append items of key asterisk to module user's privileges. This
                # potentially causes duplicate values in module values, but later step would
                # resolve such issue by converting module list to set then back to list
                user['privileges'][module_name].extend(user['privileges']['*'])

    # Loop over all modules to expand asterisk values، if any
    for module_name, module in Config.modules.items():
        if module_name not in user['privileges']:
            continue

        if '*' in user['privileges'][module_name]:
            user['privileges'][module_name].extend(module.privileges)
            # To avoid potential duplicates, convert to set, and back to list
            user['privileges'][module_name] = list(set(user['privileges'][module_name]))

    return results


async def _add_group(
    skip_events: "NawahEvents", session: "NawahSession", query: "Query", doc: "NawahDoc"
) -> "Results":
    # Check for list group attr
    if isinstance(doc["group"], list):
        for i in range(0, len(doc["group"]) - 1):
            await call(
                "user/add_group",
                skip_events=skip_events,
                session=session,
                query=query,
                doc={"group": doc["group"][i]},
            )
        doc["group"] = doc["group"][-1]
    # Confirm all basic args are provided
    doc["group"] = ObjectId(doc["group"])
    # Confirm group is valid
    results = await call(
        "group/read",
        skip_events=[Event.PERM],
        session=session,
        query=Query([{"_id": {"$eq": doc["group"]}}]),
    )

    if not results["args"]["count"]:
        raise InvalidGroupException()
    # Get user details
    results = await call(
        "user/read", skip_events=[Event.PERM], session=session, query=query
    )
    if not results["args"]["count"]:
        raise InvalidUserException()

    user = results["args"]["docs"][0]
    # Confirm group was not added before
    if doc["group"] in user["groups"]:
        raise GroupAddedException()

    user["groups"].append(doc["group"])
    # Update the user
    results = await call(
        "user/update",
        skip_events=[Event.PERM],
        session=session,
        query=query,
        doc={"groups": user["groups"]},
        args={
            "raise_no_success": True,
        },
    )
    # Check if the updated User doc belongs to current session and update it
    if session["user"]["_id"] == user["_id"]:
        user_results = await call(
            "user/read_privileges",
            skip_events=[Event.PERM],
            session=session,
            query=Query([{"_id": {"$eq": user["_id"]}}]),
        )
        session["user"] = user_results["args"]["docs"][0]

    return results


async def _delete_group(session: "NawahSession", query: "Query") -> "Results":
    # Confirm group is valid
    results = await call(
        "group/read",
        skip_events=[Event.PERM],
        session=session,
        query=Query([{"_id": {"$eq": query["group:$eq"][0]}}]),
    )

    if not results["args"]["count"]:
        raise InvalidGroupException()
    # Get user details
    results = await call(
        "user/read",
        skip_events=[Event.PERM],
        session=session,
        query=Query([{"_id": {"$eq": query["_id:$eq"][0]}}]),
    )

    if not results["args"]["count"]:
        raise InvalidUserException()

    user = results["args"]["docs"][0]
    # Confirm group was not added before
    if query["group:$eq"][0] not in user["groups"]:
        raise GroupNotAddedException()

    # Update the user
    results = await call(
        "user/update",
        skip_events=[Event.PERM],
        session=session,
        query=Query([{"_id": {"$eq": query["_id:$eq"][0]}}]),
        doc={"groups": {"$del_val": [query["group:$eq"][0]]}},
    )

    # if update fails, return update results
    if results["status"] != 200:
        return results

    # Check if the updated User doc belongs to current session and update it
    if session["user"]["_id"] == user["_id"]:
        user_results = await call(
            "user/read_privileges",
            skip_events=[Event.PERM],
            session=session,
            query=Query([{"_id": {"$eq": user["_id"]}}]),
        )
        session["user"] = user_results["args"]["docs"][0]

    return results



"""Provides enums required for Nawah Fremework and apps"""

from ._enums import (
    AttrType,
    DeleteStrategy,
    Event,
    LocaleStrategy,
    NawahValues,
    VarType,
)

__all__ = [
    "AttrType",
    "DeleteStrategy",
    "Event",
    "LocaleStrategy",
    "NawahValues",
    "VarType",
]



"""Provides 'call' Utility"""

import asyncio
import copy
import datetime
import inspect
import json
import logging
import re
import sys
import traceback
from typing import (TYPE_CHECKING, Any, Awaitable, Callable, Literal,
                    MutableMapping, MutableSequence, Type, Union, cast)

from bson import ObjectId

from nawah.classes import Query, QueryMod, Var, app_encoder
from nawah.config import Config
from nawah.enums import Event
from nawah.exceptions import (InvalidAttrException, InvalidCallArgException,
                              InvalidCallEndpointException,
                              InvalidDocAttrException, InvalidDocException,
                              InvalidFuncException, InvalidModuleException,
                              InvalidQueryAttrException, InvalidQueryException,
                              MissingDocAttrException,
                              MissingQueryAttrException,
                              UnknownCallArgException)

from ._cache import _get_cache, _set_cache, reset_cache_channel
from ._check_permissions import check_permissions as check_permissions_utility
from ._val import var_value
from ._validate import validate_attr

if TYPE_CHECKING:
    from nawah.classes import Attr
    from nawah.types import NawahDoc, NawahEvents, NawahSession, Results


logger = logging.getLogger("nawah")


async def call(
    endpoint: str,
    /,
    *,
    module_name: str = None,
    skip_events: "NawahEvents" = None,
    env=None,
    session: "NawahSession" = None,
    query: "Query" = None,
    doc: "NawahDoc" = None,
    args: MutableMapping[str, Any] = None,
) -> "Results":
    """Checks validity of an endpoint and calls Nawah Function callable at endpoint,
    returning the couroutine of the callable. If endpoint points to non-existent
    Nawah Function, raises 'InvalidFuncException'"""

    if not re.match(r"^[a-z_]+\/[a-z_]+$", endpoint):
        raise InvalidCallEndpointException(endpoint=endpoint)

    endpoint_module, endpoint_func = endpoint.split("/")

    try:
        module = Config.modules[endpoint_module]
    except KeyError as e:
        raise InvalidModuleException(module_name=endpoint_module) from e

    try:
        func = module.funcs[endpoint_func]
    except KeyError as e:
        raise InvalidFuncException(
            module_name=endpoint_module, func_name=endpoint_func
        ) from e

    # Set defaults for kwargs
    module_name = module_name or endpoint_module
    skip_events = skip_events or []
    query = query or Query()
    doc = doc or {}
    if not session:
        if env:
            session = env["session"]
        else:
            # Deep-copy system session to prevent manipulating it, but skip conn as it can't be
            # deep-copied
            session = copy.deepcopy(
                {k: v for k, v in Config.sys.session.items() if k != "conn"}
            )
            # Then add it again
            session['conn'] = Config.sys.session['conn']
    args = args or {}

    # call Utility could be used from app handlers directly, where value for query would be
    # Query-compatible dict. Refactor onto Query
    if isinstance(query, dict) and ("$pipe" in query or "$special" in query):
        if "$pipe" not in query:
            query["$pipe"] = []
        if "$special" not in query:
            query["$special"] = {}

        query = Query(query["$pipe"], special=query["$special"])

    # [TODO] Remove when legacy query format is obsoleted
    # Convert query to Query object
    if not isinstance(query, Query):
        logger.warning("Call is using deprecated query format: %s", query)
        query_special = {}
        for step in query:
            if not isinstance(step, dict):
                continue
            for attr_name, attr_val in step.items():
                if isinstance(attr_name, str) and attr_name[0] == "$":
                    query_special[attr_name] = attr_val
        query = Query(_convert_deprecated_query(query), special=query_special)

    # Check conditions for call checks
    check_permissions = Event.PERM not in skip_events
    check_attrs_query = Event.ATTRS_QUERY not in skip_events
    check_attrs_doc = Event.ATTRS_DOC not in skip_events
    check_cache = Event.CACHE not in skip_events
    check_analytics = Event.ANALYTICS not in skip_events

    if check_permissions:
        query_mod, doc_mod = check_permissions_utility(func=func, session=session)
        if query_mod and not isinstance(query_mod, QueryMod):
            logger.warning(
                "Evaluated permission Set Query Modifier is using deprecated query format: %s",
                query,
            )
            if isinstance(query_mod, dict):
                query_mod = [query_mod]
            query_special = {}
            for step in query_mod:
                if not isinstance(step, dict):
                    continue
                for attr_name, attr_val in step.items():
                    if isinstance(attr_name, str) and attr_name[0] == "$":
                        query.special[attr_name] = attr_val
            query_mod = QueryMod(_convert_deprecated_query(query_mod))

        # Use return of check_permissions_utility to update query, doc
        if query_mod:
            for step in query_mod.pipe:
                query.append(
                    _process_query_mod(
                        query=query,
                        query_mod=cast(MutableMapping, step),
                        session=session,
                        doc=doc,
                    )
                )
            query.special.update(query_mod.special)

        if doc_mod:
            doc.update(
                _process_doc_mod(
                    doc=doc,
                    doc_mod=doc_mod,
                    session=session,
                    query=query,
                )
            )

    if check_attrs_query:
        # Sanitise $deleted Special Query Attr, if ATTRS_QUERY Event is not skipped
        if "$deleted" in query:
            del query["$deleted"]
        func.query_attrs = cast(MutableSequence, func.query_attrs)
        _check_query_attrs(query=query, query_attrs=func.query_attrs)

    if check_attrs_doc:
        func.doc_attrs = cast(MutableSequence, func.doc_attrs)
        _check_doc_attrs(doc=doc, doc_attrs=func.doc_attrs)

    if check_cache:
        cache_key, call_cache = await _get_cache(
            func=func,
            skip_events=skip_events,
            session=session,
            query=query,
        )

        if call_cache:
            return call_cache

    if check_analytics:
        # [TODO] Implement
        pass

    try:
        func_callable = cast(Callable[..., Awaitable["Results"]], func.callable)
        kwargs: MutableMapping = {
            "func": func,
            "module_name": module_name,
            "skip_events": skip_events,
            "env": {"session": session},
            "session": session,
            "query": query,
            "doc": doc,
        }

        # Iterate over call_args to set values
        for arg_name, arg_type in (func.call_args or {}).items():
            if arg_name not in args:
                kwargs[arg_name] = None
                continue

            try:
                validate_attr(
                    mode="update",
                    attr_name=arg_name,
                    attr_type=arg_type,
                    attr_val=args[arg_name],
                )
            except InvalidAttrException as e:
                raise InvalidCallArgException(
                    module_name=module_name, func_name=func.name, arg_name=arg_name
                ) from e
            # Call Arg value is value, append it to kwargs to be passed to Func callable
            kwargs[arg_name] = args[arg_name]
            # Delete arg_val from args to filter out invalid args
            del args[arg_name]

        # If after iterating over call_args, we still have values in args, declare them invalid
        if args:
            arg_name = list(args)[0]
            raise UnknownCallArgException(
                module_name=module_name, func_name=func.name, arg_name=arg_name
            )

        results = await func_callable(
            **{
                param: kwargs[param]
                for param in inspect.signature(func_callable).parameters
                if param in kwargs
            }
        )
    except Exception as e:  # pylint: disable=broad-except
        # Handling exceptions raised in callables goes in one of four conditions

        # No value, or empty dict, for Func.exceptions, such as no exceptions at all expected
        # Raise with reason=no_exceptions, will report as error
        if not func.exceptions:
            _log_and_raise(e=e, endpoint=endpoint, reason="no_exceptions")

        func.exceptions = cast(MutableMapping[Type[Exception], bool], func.exceptions)

        # Exception not in Func.exceptions, such as exception is not expected
        # Raise with reason=exception_not_in, will report as error
        if type(e) not in func.exceptions:
            _log_and_raise(e=e, endpoint=endpoint, reason="exception_not_in")

        # Exception in Func.exceptions, and set to True, such as exception is expected, but still
        # report as error
        # Raise with reason=exception_true, will report as error
        if func.exceptions[type(e)]:
            _log_and_raise(e=e, endpoint=endpoint, reason="exception_true")

        # Exception in Func.exceptions, and set to False, such as exception is expected, no report
        # required. Only scenario as such
        # Raise with reason=exception_false, will log details only
        _log_and_raise(e=e, endpoint=endpoint, reason="exception_false")

    if "args" in results and "session" in results["args"]:
        if "session" == ObjectId("f00000000000000000000012"):
            from ._config import _compile_anon_session, _compile_anon_user

            anon_session = _compile_anon_session()
            anon_session["user"] = _compile_anon_user()
            session.update(anon_session)
        else:
            session.update(json.loads(app_encoder.encode(results["args"]["session"])))

    if check_cache and cache_key:
        results["args"]["cache_key"] = cache_key
        if "cache_time" not in results["args"]:
            logger.debug("Results generated with 'cache_key'. Calling '_set_cache'.")
            results["args"]["cache_time"] = datetime.datetime.utcnow().isoformat()
        # [TODO] Add callback to handle errors
        asyncio.create_task(_set_cache(func=func, cache_key=cache_key, results=results))

    if func.cache_channels_reset:
        for channel in func.cache_channels_reset:
            # [TODO] Add callback to handle errors
            asyncio.create_task(reset_cache_channel(channel))

    return results


def _log_and_raise(
    *,
    e: Exception,
    endpoint: str,
    reason: Literal[
        "no_exceptions", "exception_true", "exception_false", "exception_not_in"
    ],
):
    exc_type, exc_value, exc_traceback = sys.exc_info()
    logging_method = logger.error
    if reason == "exception_false":
        logging_method = logger.error

    logging_method(
        "Callable for '%s' failed with Exception of type: %s", endpoint, type(e)
    )

    logging_method("Exception traceback:")
    for line in traceback.format_exception(exc_type, exc_value, exc_traceback):
        logging_method("- %s", line)

    tb = cast(Any, exc_traceback)

    if tb.tb_next:
        logging_method("Exception frame locals:")
        logging_method(tb.tb_next.tb_frame.f_locals)

    raise e


def _process_query_mod(
    *,
    query_mod: Union[
        MutableMapping[str, Any], MutableSequence[MutableMapping[str, Any]]
    ],
    session: "NawahSession",
    query: "Query",
    doc: "NawahDoc",
):
    if not query_mod:
        return {}

    query_mod_processed: Union[
        MutableMapping[str, Any], MutableSequence[MutableMapping[str, Any]]
    ]

    if isinstance(query_mod, list):
        query_mod_processed = []
        for query_mod_child in query_mod:
            query_mod_processed.append(
                _process_query_mod(
                    query_mod=query_mod_child,
                    session=session,
                    query=query,
                    doc=doc,
                )
            )

    elif isinstance(query_mod, dict):
        query_mod_processed = {}
        for attr_name, attr_val in query_mod.items():
            if isinstance(attr_val, (list, dict)):
                query_mod_processed[attr_name] = _process_query_mod(
                    query_mod=attr_val,
                    session=session,
                    query=query,
                    doc=doc,
                )
            elif isinstance(attr_val, Var):
                query_mod_processed[attr_name] = var_value(
                    attr_val,
                    session=session,
                    doc=doc,
                    locale=session["user"]["locale"],
                )
            else:
                query_mod_processed[attr_name] = copy.deepcopy(attr_val)
    else:
        raise Exception("Unexpected non-list, non-dict value for 'query_mod'")

    return query_mod_processed


def _process_doc_mod(
    *,
    doc_mod: Union[MutableMapping[str, Any], MutableSequence[MutableMapping[str, Any]]],
    session: "NawahSession",
    query: "Query",
    doc: "NawahDoc",
):
    if not doc_mod:
        return {}

    doc_mod_processed: Union[
        MutableMapping[str, Any], MutableSequence[MutableMapping[str, Any]]
    ]

    if isinstance(doc_mod, list):
        doc_mod_processed = []
        for doc_mod_child in doc_mod:
            doc_mod_processed.append(
                _process_doc_mod(
                    doc_mod=doc_mod_child,
                    session=session,
                    query=query,
                    doc=doc,
                )
            )

    elif isinstance(doc_mod, dict):
        doc_mod_processed = {}
        for attr_name, attr_val in doc_mod.items():
            if isinstance(attr_val, (list, dict)):
                doc_mod_processed[attr_name] = _process_doc_mod(
                    doc_mod=attr_val,
                    session=session,
                    query=query,
                    doc=doc,
                )
            elif isinstance(attr_val, Var):
                doc_mod_processed[attr_name] = var_value(
                    attr_val,
                    session=session,
                    doc=doc,
                    locale=session["user"]["locale"],
                )
            else:
                doc_mod_processed[attr_name] = copy.deepcopy(attr_val)
    else:
        raise Exception("Unexpected non-list, non-dict value for 'doc_mod'")

    return doc_mod_processed


def _check_query_attrs(
    *,
    query: "Query",
    query_attrs: MutableSequence[MutableMapping[str, "Attr"]],
):
    if not query_attrs:
        return

    query_attrs_sets: MutableSequence[MutableMapping[str, str]] = []

    for query_attrs_set in query_attrs:
        query_attrs_sets.append({})
        try:
            for attr_name, attr_type in query_attrs_set.items():
                # [TODO] Remove when nawah query is obsoleted
                if ":" not in attr_name:
                    logger.warning(
                        "Detected attr for Func.query_attrs with no oper: %s", attr_name
                    )
                    attr_name = f"{attr_name}:$eq"
                if attr_name not in query:
                    query_attrs_sets[-1][attr_name] = "missing"
                    raise MissingQueryAttrException(attr_name=attr_name)

                for i, attr_val in enumerate(query[attr_name]):
                    try:
                        query[attr_name][i] = validate_attr(
                            mode="create",
                            attr_name=attr_name,
                            attr_type=attr_type,
                            attr_val=attr_val,
                        )
                    except InvalidAttrException as e:
                        query_attrs_sets[-1][attr_name] = "invalid"
                        raise InvalidQueryAttrException(
                            attr_name=attr_name,
                            attr_type=attr_type,
                            val_type=type(attr_val),
                        ) from e

                query_attrs_sets[-1][attr_name] = "valid"
            # If looped successfully over complete set, return to indicate valid Query
            return
        except (InvalidAttrException, MissingQueryAttrException):
            # If exception occur, pass it to allow checking all sets
            pass

    # If all sets are checked but failed to return, rase InvalidQueryException
    raise InvalidQueryException(query_attrs_sets=query_attrs_sets)


def _check_doc_attrs(
    *,
    doc: "NawahDoc",
    doc_attrs: MutableSequence[MutableMapping[str, "Attr"]],
):
    if not doc_attrs:
        return

    doc_attrs_sets: MutableSequence[MutableMapping[str, str]] = []

    for doc_attrs_set in doc_attrs:
        doc_attrs_sets.append({})
        try:
            for attr_name, attr_type in doc_attrs_set.items():
                if attr_name not in doc:
                    doc_attrs_sets[-1][attr_name] = "missing"
                    raise MissingDocAttrException(attr_name=attr_name)

                try:
                    doc[attr_name] = validate_attr(
                        mode="create",
                        attr_name=attr_name,
                        attr_type=attr_type,
                        attr_val=doc[attr_name],
                    )
                except InvalidAttrException as e:
                    doc_attrs_sets[-1][attr_name] = "invalid"
                    raise InvalidDocAttrException(
                        attr_name=attr_name,
                        attr_type=attr_type,
                        val_type=type(doc[attr_name]),
                    ) from e

                doc_attrs_sets[-1][attr_name] = "valid"
            # If looped successfully over complete set, return to indicate valid Doc
            return
        except (InvalidAttrException, MissingDocAttrException):
            # If exception occur, pass it to allow checking all sets
            pass

    # If all sets are checked but failed to return, rase InvalidDocException
    raise InvalidDocException(doc_attrs_sets=doc_attrs_sets)


def _convert_deprecated_query(query):
    query = copy.deepcopy(query)
    shadow_query = []
    for step in query:
        if isinstance(step, list):
            shadow_query.append({"$or": _convert_deprecated_query(step)})

        elif isinstance(step, dict):
            for attr, val in step.items():
                if attr[0] == "$":
                    continue
                if isinstance(val, dict) and (
                    len(val) and list(val)[0].startswith("$")
                ):
                    shadow_step = {attr: val}
                else:
                    shadow_step = {attr: {"$eq": val}}

                if attr == "_id":
                    if isinstance(
                        shadow_step["_id"][list(shadow_step["_id"])[0]], list
                    ):
                        shadow_step["_id"][list(shadow_step["_id"])[0]] = [
                            ObjectId(item)
                            for item in shadow_step["_id"][list(shadow_step["_id"])[0]]
                        ]
                    else:
                        shadow_step["_id"][list(shadow_step["_id"])[0]] = ObjectId(
                            shadow_step["_id"][list(shadow_step["_id"])[0]]
                        )

                shadow_query.append(shadow_step)

    return shadow_query



"""Provides utilities required for configuring modules, packages and app"""

import copy
import datetime
import logging
import os
import re
from dataclasses import fields
from typing import TYPE_CHECKING, Any, MutableMapping, MutableSequence, cast

import sentry_sdk
from bson import ObjectId
from croniter import croniter
from passlib.hash import pbkdf2_sha512
from redis import asyncio as aioredis  # type: ignore

from nawah.classes import (App, Attr, Counter, Default, Diff, Extn, Func, Perm,
                           SysDoc, UserAttr, Var)
from nawah.config import Config
from nawah.enums import AttrType, Event
from nawah.exceptions import (ConfigException, InvalidAttrTypeArgException,
                              InvalidAttrTypeException, InvalidVarException)

from ._attr_type import extract_attr, generate_attr_val
from ._call import call
from ._val import deep_update, var_value
from ._validate import validate_attr, validate_type

if TYPE_CHECKING:
    from nawah.classes import Env, Module, Package
    from nawah.types import NawahDoc, NawahSession

logger = logging.getLogger("nawah")


def process_env_vars(config: "Env", /):
    """Checks all Var-prune Config Attrs of 'config' and updates value"""

    var_args: MutableMapping[str, MutableMapping[str, Any]] = {
        "debug": {"var": config.debug, "type": bool, "default": False},
        "port": {"var": config.port, "type": int, "default": 8081},
        "conn_timeout": {"var": config.conn_timeout, "type": int},
        "quota_anon_min": {"var": config.quota_anon_min, "type": int},
        "quota_auth_min": {"var": config.quota_auth_min, "type": int},
        "quota_ip_min": {"var": config.quota_ip_min, "type": int},
        "file_upload_limit": {
            "var": config.file_upload_limit,
            "type": int,
        },
        "file_upload_timeout": {
            "var": config.file_upload_timeout,
            "type": int,
        },
        "data_server": {"var": config.data_server, "type": str},
        "data_name": {"var": config.data_name, "type": str},
        "data_ssl": {"var": config.data_ssl, "type": bool},
        "data_disk_use": {"var": config.data_disk_use, "type": bool},
        "cache_server": {"var": config.cache_server, "type": str},
        "cache_db": {
            "var": config.cache_db,
            "type": int,
            "default": 0,
        },
        "cache_username": {"var": config.cache_username, "type": str},
        "cache_password": {"var": config.cache_password, "type": str},
        "cache_expiry": {"var": config.cache_expiry, "type": int},
        "error_reporting_server": {
            "var": config.error_reporting_server,
            "type": str,
        },
    }

    for arg_name, var_arg in var_args.items():
        if not isinstance(var_arg["var"], Var):
            continue
        try:
            setattr(
                config, arg_name, var_value(var_arg["var"], return_type=var_arg["type"])
            )
        except InvalidVarException as e:
            if "default" not in var_arg:
                raise e
            setattr(config, arg_name, var_arg["default"])


def _pre_setup_app(app_config: "App", /):
    # Check envs, env App Config Attr
    if app_config.envs and app_config.env:
        if isinstance(app_config.env, Var):
            logger.debug(
                "App '{name}' defines 'env' App Config Attr as 'Var' object. Getting its "
                "value.."
            )
            app_config.env = var_value(app_config.env)

        if app_config.env not in app_config.envs:
            raise ConfigException(
                "Value for 'env' App Config Attr is '{app_config.env}', but"
                "no such Env is defined"
            )

        # Update runtime config with values from Env object
        app_config.env = cast(str, app_config.env)
        env = app_config.envs[app_config.env]
        for env_field in fields(env):
            if (env_attr_val := getattr(env, env_field.name)) is not None:
                # Specifically for l10n Config Attr, always update base value, and don't overwrite it
                if env_field.name == "l10n":
                    # For a possibility app_config.l10n is None, set as L10N
                    if app_config.l10n is None:
                        app_config.l10n = {}
                    app_config.l10n.update(env_attr_val)
                else:
                    setattr(app_config, env_field.name, env_attr_val)

    if app_config.debug is True:
        logger.setLevel(logging.DEBUG)
        logger.debug("Set logging level to DEBUG (Config.debug==True)")

    if isinstance(app_config.debug, Var):
        try:
            if var_value(app_config.debug, return_type=lambda _: bool(int(_))):
                logger.setLevel(logging.DEBUG)
                logger.debug("Set logging level to DEBUG (Config.debug.var)")
                app_config.debug = True
        except InvalidVarException:
            logger.warning(
                "No value for 'Var' object for 'debug' App Config Attr. Logging "
                "remains at level ERROR"
            )

    logger.debug("Getting value of 'Var' object of 'admin_password' App Config Attr..")
    if isinstance(app_config.admin_password, Var):
        try:
            admin_password_val = var_value(app_config.admin_password)
            if not re.match(
                r"^(?=.*[a-z])(?=.*[A-Z])(?=.*[0-9])(?=.{8,})", admin_password_val
            ):
                raise Exception(
                    f"Value '{admin_password_val}' of 'admin_password' App Config Attr doesn't "
                    "match accepted password format"
                )
            app_config.admin_password = admin_password_val
            Config.sys.admin_check = True
        except InvalidVarException:
            logger.warning(
                "No value for 'Var' object for 'admin_password' App Config Attr. 'ADMIN' doc "
                "will not be checked or updated"
            )
            Config.sys.admin_check = False

    if isinstance(app_config.anon_token, Var):
        logger.debug(
            "App '{name}' defines 'anon_token' App Config Attr as 'Var' object. Getting its "
            "value.."
        )
        app_config.anon_token = var_value(app_config.anon_token)

    process_env_vars(app_config)


def setup_app(app_config: "App", /, *, process_vars: bool = False):
    # pylint: disable=import-outside-toplevel

    import nawah.base as Base
    from nawah.packages.core import core

    app_config_version = cast(str, app_config.version)

    if not re.match(r"^[0-9]+\.[0-9]+\.[0-9]+([ab][0-9]+)?$", app_config_version):
        raise ConfigException(
            f"App '{app_config.name}' defines invalid 'version' '{app_config.version}'"
        )

    if process_vars:
        _pre_setup_app(app_config)

    configs: MutableSequence["Package"] = [
        core,
        *(app_config.packages or []),
        app_config,
    ]

    for config in configs:
        for config_field in fields(config):
            attr_name = config_field.name
            attr_val = getattr(config, attr_name, None)

            if attr_val is None:
                continue

            if attr_name in ["envs", "api_level", "version", "modules", "packages"]:
                continue

            # For user_attrs, process separatly to update user module
            if attr_name == "user_attrs":
                user_module = Config.modules["user"]

                # Loop over user_attrs to validate and process
                for user_attr_name, user_attr in cast(
                    MutableMapping[str, "UserAttr"], config.user_attrs
                ).items():
                    # Validate value type
                    if not isinstance(user_attr, UserAttr):
                        raise ConfigException(
                            f"Invalid type for attr '{user_attr_name}' defined for package "
                            f"'{config.name}' for 'user_attrs' Config Attr"
                        )
                    # Make sure no conflicts with current attrs in user happen
                    if user_attr_name in user_module.attrs:
                        raise ConfigException(
                            f"Package '{config.name}' defines 'user_attrs' Config Attr with "
                            f"conflicting attr '{user_attr_name}'"
                        )
                    # And, validate UserAttr.type
                    if not isinstance(user_attr.type, Attr):
                        raise ConfigException(
                            f"Invalid type for attr type '{user_attr_name}' defined for package "
                            f"'{config.name}' for 'user_attrs' Config Attr"
                        )

                    # Things are good. Add attr to both user_attrs Config Attr, which tracks attrs
                    # added to 'user' Nawah Module with runtime config, and to 'user' attrs
                    Config.user_attrs[user_attr_name] = user_attr
                    user_module.attrs[user_attr_name] = user_attr.type

                    # Add attr to user_attrs_sanitise Config Attr if defined as such. This makes
                    # calls to 'user/read' endpoint without 'skip_sanitise_results' Call Arg
                    # removes references values from user doc
                    if user_attr.sanitise:
                        Config.user_attrs_sanitise.append(user_attr_name)

                    # If UserAttr is defined as auth_attr, Update all values from 'user', 'session'
                    # Nawah Modules which manage and control authentication and unique values for
                    # auth_attrs
                    if user_attr.auth_attr:
                        # Raise ConfigException if UserAttr defines both auth_attr, and default,
                        # Ultimately, allowing creating 'user' doc with NoneType-value for attr,
                        # but fails with DuplicateUniqueException
                        if user_attr.default:
                            raise ConfigException(
                                "Invalid UserAttr defining 'auth_attr', 'default' for attr "
                                f"'{user_attr_name}' defined for package '{config.name}' for "
                                "'user_attrs' Config Attr"
                            )

                        Config.auth_attrs.append(user_attr_name)
                        Config.user_attrs_sanitise.append(f"{user_attr_name}_hash")
                        user_module.unique_attrs.append(user_attr_name)
                        user_module.attrs[f"{user_attr_name}_hash"] = Attr.STR()
                        session_auth_doc_args = cast(
                            MutableSequence[MutableMapping[str, "Attr"]],
                            Config.modules["session"].funcs["auth"].doc_attrs,
                        )
                        session_auth_doc_args.extend(
                            [
                                {
                                    "hash": Attr.STR(),
                                    user_attr_name: user_attr.type,
                                    "groups": Attr.LIST(list=[Attr.ID()]),
                                },
                                {"hash": Attr.STR(), user_attr_name: user_attr.type},
                            ]
                        )

                    # If UserAttr defines a default, validate and update
                    if user_attr.default:
                        if not isinstance(user_attr.default, Default):
                            raise ConfigException(
                                f"Invalid type for attr default '{user_attr_name}' defined for "
                                f"package '{config.name}' for 'user_attrs' Config Attr"
                            )
                        # Since Nawah Module 'defaults' attribute is config approach to setting
                        # 'default' attribute of respectful Attr objects. By this point, 'defaults'
                        # for 'user' are already processed, so 'default' attribute of UserAttr.type
                        # must be updated manually
                        user_module.defaults[user_attr_name] = user_attr.default
                        user_attr.type.default = user_attr.default

                    # If UserAttr defines extension instruction, validate and update
                    if user_attr.extn:
                        if not isinstance(user_attr.extn, Extn):
                            raise ConfigException(
                                f"Invalid type for attr extn '{user_attr_name}' defined for "
                                f"package '{config.name}' for 'user_attrs' Config Attr"
                            )
                        # Similar with manually needing to update Attr object 'extn' attribute
                        user_module.extns[user_attr_name] = user_attr.extn
                        user_attr.type.extn = user_attr.extn

                    if user_attr.func:
                        # [TODO] Expanded type validation for Func
                        if not isinstance(user_attr.type, Func):
                            raise ConfigException(
                                f"Invalid type for attr func '{user_attr_name}' defined for "
                                f"package '{config.name}' for 'user_attrs' Config Attr"
                            )
                        user_attr_func = user_module.funcs[
                            f"update_{user_attr_name}"
                        ] = user_attr.func
                    else:
                        user_attr_func = user_module.funcs[
                            f"update_{user_attr_name}"
                        ] = Func(
                            permissions=[Perm(privilege=user_attr_name)],
                            query_attrs={"_id": Attr.ID()},
                            doc_attrs={user_attr_name: user_attr.type},
                            strict_doc=True,
                        )

                    user_attr_func.name = f"update_{user_attr_name}"
                    user_attr_func.module = user_module
                    user_attr_func.call_args = {"raise_no_success": Attr.BOOL()}
                    user_attr_func.callable = Base.update

            # For vars_types Config Attr, preserve config name for debugging purposes
            elif attr_name == "vars_types":
                for var in attr_val:
                    config.name = cast(str, config.name)
                    Config.vars_types[var] = {
                        "package": config.name,
                        "type": attr_val[var],
                    }

            # Otherwise, update Config accroding to attr_val type
            elif isinstance(attr_val, list):
                for j in attr_val:
                    getattr(Config, attr_name).append(j)
                if attr_name == "locales":
                    Config.locales = list(set(Config.locales))

            elif isinstance(attr_val, dict):
                if not getattr(Config, attr_name):
                    setattr(Config, attr_name, {})
                deep_update(target=getattr(Config, attr_name), new_values=attr_val)

            else:
                setattr(Config, attr_name, attr_val)

        if config.modules:
            config_modules = []
            for module in config.modules:
                if module.name in Config.modules:
                    raise ConfigException(
                        f"Module '{module.name}' exist in current runtime config"
                    )

                config_module(module_name=module.name, module=module)
                config_modules.append(module.name)
                Config.modules[module.name] = module

            Config.sys.packages[config.name or ""] = {
                "package": config,
                "modules": config_modules,
            }

        if getattr(config, "packages", None):
            pass

        if isinstance(config, App):
            Config.sys.name = config.name or ""
            Config.sys.version = config.version or ""

    # Update User, Session modules with populated attrs
    user_module = Config.modules["user"]
    user_locale_default = Default(value=Config.locale)
    user_module.defaults["locale"] = user_locale_default
    user_module.attrs["locale"].default = user_locale_default

    # [DOC] Attempt to validate all packages required vars (via vars_types Config Attr) are met
    for var_name, var in Config.vars_types.items():
        if var_name not in Config.vars:
            raise ConfigException(
                f'Package \'{var["package"]}\' requires \'{var_name}\' Var, but not found in App '
                "Config"
            )
        try:
            validate_attr(
                mode="create",
                attr_name=var_name,
                attr_type=var["type"],
                attr_val=Config.vars[var_name],
            )
        except Exception as e:
            raise ConfigException(
                f'Package \'{var["package"]}\' requires \'{var_name}\' Var of type '
                f'\'{var["type"].type}\', but validation failed'
            ) from e

    # Validate Attr Type of type TYPE. Has to run before config_app so calls in config_app
    # depending on such attrs don't fail
    for attr_type in Config.sys.type_attrs:
        validate_type(attr_type=attr_type)

    # Attempt to call post_config callable of all modules
    for module_name, module in Config.modules.items():
        logger.debug(
            "Attempting to check for 'post_config' callable for module: %s",
            module_name,
        )
        if not module.post_config:
            logger.debug("- Module does not define 'post_config' callable")
            continue

        logger.debug("- Module does define 'post_config' callable. Calling it..")
        module.post_config()


def config_module(*, module_name: str, module: "Module"):
    """Validates args for 'module' and updates values for 'attrs' of 'module' per sets from
    'defaults', 'extns', 'counters'"""

    # pylint: disable=import-outside-toplevel

    import nawah.base as Base

    # Set name of module in Module object for reverse reference
    module.name = module_name

    logger.debug("Attempting to configure module '%s'", module_name)

    if not module.funcs:
        raise ConfigException(f"Module '{module_name}' doesn't define any function")

    for func_name, func in module.funcs.items():
        logger.debug(
            "Attempting to validate Function '%s'.'%s'", module_name, func_name
        )
        func.name = func_name
        func.module = module
        # Check value type
        if not isinstance(func, Func):
            raise ConfigException(
                f"Invalid Nawah Function for '{module_name}'.'{func_name}' of type "
                "'{type(func)}'"
            )

        # Validate query_attrs, doc_attrs, call_args
        for func_attr_name in ["query_attrs", "doc_attrs", "call_args"]:
            logger.debug(
                "Attempting to validate '%s' of '%s'.'%s'",
                func_attr_name,
                module_name,
                func_name,
            )
            if func_attr_val := getattr(func, func_attr_name):
                # query_attrs, doc_attrs are lists of dicts, call_args is just dict
                # Do this trick to get this loop to work nonetheless, by wrapping call_args with
                # a list, to let the rest of the loop deal with theree attrs in same manner
                if func_attr_name == "call_args":
                    func_attr_val = [func_attr_val]
                for attrs_set in func_attr_val:
                    for attr_name, attr_type in attrs_set.items():
                        try:
                            logger.debug(
                                "Attempting to validate '%s' Attr Type of '%s' for Function "
                                "'%s'.'%s' of set '%s'",
                                func_attr_name,
                                attr_name,
                                module_name,
                                func.name,
                                attrs_set,
                            )
                            validate_type(attr_type=attr_type, skip_type=True)
                        except InvalidAttrTypeException as e:
                            raise ConfigException(
                                f"Invalid Attr Type for '{module_name}'.'{attr_name}'. "
                                f"Original validation error: {str(e)}",
                            ) from e
                        except InvalidAttrTypeArgException as e:
                            raise ConfigException(
                                f"Invalid Attr Type Arg for '{module_name}'.'{attr_name}'. "
                                f"Original validation error: {str(e)}"
                            ) from e

        # [TODO] Add steps to update Func object attributes per func callable
        # If Func has callable defined skip
        if func.callable:
            if not callable(func.callable):
                raise ConfigException(
                    f"Value 'callable' for Nawah Function '{func_name}' of '{module_name}' "
                    "is not callable"
                )
            continue

        logger.debug(
            "Function '%s'.'%s' doesn't define callable. Attempting to set per name",
            module_name,
            func_name,
        )
        # Otherwise, set callable based on fucntion name, update query_attrs, doc_attrs, call_args
        if func_name.startswith("create_file"):
            func.doc_attrs = [
                {
                    "attr": Attr.STR(),
                    "lastModified": Attr.INT(),
                    "type": Attr.STR(),
                    "name": Attr.STR(),
                    "content": Attr(
                        desc="",
                        type=AttrType.BYTES,
                        args={},
                    ),
                },
                {"_id": Attr.ID(), "doc": Attr.ID()},
            ]
            func.post_func = True
            func.callable = Base.create_file

        elif func_name.startswith("update_file"):
            func.post_func = True
            func.callable = Base.update_file

        elif func_name.startswith("delete_file"):
            func.post_func = True
            func.callable = Base.delete_file

        elif func_name.startswith("delete_lock"):
            func.callable = Base.delete_lock

        elif func_name.startswith("obtain_lock"):
            func.callable = Base.obtain_lock

        elif func_name.startswith("read_diff"):
            func.callable = Base.read_diff

        elif func_name.startswith("retrieve_file"):
            func.query_attrs = [
                {
                    "_id": Attr.ID(),
                    "attr": Attr.STR(),
                    "filename": Attr.STR(),
                },
                {
                    "_id": Attr.ID(),
                    "attr": Attr.STR(),
                    "thumb": Attr.STR(pattern=r"[0-9]+x[0-9]+"),
                    "filename": Attr.STR(),
                },
            ]
            func.get_func = True
            func.callable = Base.retrieve_file

        elif func_name.startswith("read"):
            func.callable = Base.read

        elif func_name.startswith("create"):
            func.callable = Base.create

        elif func_name.startswith("update"):
            func.callable = Base.update

        elif func_name.startswith("delete"):
            func.callable = Base.delete

        else:
            raise ConfigException(
                f"Can't set callable for Nawah Function '{func_name}' of"
                f" '{module_name}'"
            )

        if not func.call_args:
            func.call_args = {}

        func.call_args.update({"raise_no_success": Attr.BOOL()})

    # Check attrs for any invalid type
    for attr in module.attrs:
        try:
            logger.debug(
                "Attempting to validate Attr Type for '%s'.'%s'",
                module_name,
                attr,
            )
            validate_type(attr_type=module.attrs[attr], skip_type=True)
        except InvalidAttrTypeException as e:
            raise ConfigException(
                f"Invalid Attr Type for '{module_name}'.'{attr}'. Original "
                f"validation error: {str(e)}",
            ) from e
        except InvalidAttrTypeArgException as e:
            raise ConfigException(
                f"Invalid Attr Type Arg for '{module_name}'.'{attr}'. Original "
                f"validation error: {str(e)}"
            ) from e

    # Check defaults for invalid types, update default value
    for attr_name, default in module.defaults.items():
        if not isinstance(default, Default):
            raise ConfigException(
                f"Invalid Default Instruction for '{module_name}'.'{attr_name}', of type "
                f"'{type(default)}'"
            )

        logger.debug(
            "Updating default value for attr '%s' to: '%s'",
            attr_name,
            module.defaults[attr_name],
        )
        extract_attr(attrs=module.attrs, path=attr_name).default = module.defaults[
            attr_name
        ]

    # Check counters for invalid attrs, types
    for attr_name, counter in module.counters.items():
        if not isinstance(counter, Counter):
            raise ConfigException(
                f"Invalid Counter Instruction for '{module_name}'.'{attr_name}', of type "
                f"'{type(counter)}'"
            )

        if attr_name not in module.attrs:
            raise ConfigException(
                f"Invalid attr for Counter Instruction. Module '{module_name}' "
                f"doesn't have top-level attr '{attr_name}'"
            )

        if module.attrs[attr_name].type != AttrType.ANY:
            raise ConfigException(
                f"Attr '{module_name}'.'{attr_name}' for Counter Instruction can only be of "
                "Attr Type ANY"
            )

        # Set default value to Default(value=None) to skip InvalidAttrException when checking doc
        module.attrs[attr_name].default = Default(value=None)

    # Update extn value
    for attr_name, extn in module.extns.items():
        if not isinstance(extn, Extn):
            raise ConfigException(
                f"Invalid Extension Instruction for '{module_name}'.'{attr_name}', of type "
                f"'{type(extn)}'"
            )

        logger.debug(
            "Updating 'extn' for attr '%s' to: '%s'",
            attr_name,
            module.extns[attr_name],
        )
        extract_attr(attrs=module.attrs, path=attr_name).extn = module.extns[attr_name]

    # Check extns for invalid extended attrs
    for attr_name, extn in module.extns.items():
        if not isinstance(extn, Extn):
            raise ConfigException(
                f"Invalid extns attr '{attr_name}' of module '{module_name}'"
            )

    # Check valid type, value for diff
    if module.diff is not None and not isinstance(module.diff, Diff):
        raise ConfigException(f"Invalid Diff Instruction for module '{module_name}'")

    if isinstance(module.diff, Attr):
        module.diff = cast(Attr, module.diff)
        if module.diff.type != AttrType.TYPE:
            raise ConfigException(
                f"Invalid Attr Type for diff of module '{module_name}'. Only Attr Type TYPE "
                "is allowed"
            )
        logger.debug(
            "Attempting to validate Attr Type diff of module '%s'", module_name
        )
        try:
            validate_type(attr_type=module.diff)
        except InvalidAttrTypeException as e:
            raise ConfigException(
                f"Invalid Attr Type for diff of module '{module_name}'. Original validation error: {str(e)}"
            ) from e
        except InvalidAttrTypeArgException as e:
            raise ConfigException(
                f"Invalid Attr Type Arg for diff of module '{module_name}'. Original validation error: {str(e)}"
            ) from e

    # [TODO] Re-implement
    # Check valid types, values for create_draft, update_draft
    # for attr in ['create_draft', 'update_draft']:
    #     if type(getattr(self, attr)) not in [bool, Attr]:
    #         raise ConfigException(f'Invalid {attr} for module \'{module_name}\'') from e
    #     if type(getattr(self, attr)) == Attr:
    #         if getattr(self, attr).type != AttrType.TYPE:
    #             raise ConfigException(
    #                 f'Invalid Attr Type for {attr} of module \'{module_name}\'. Only Attr Type TYPE is allowed'
    #             ) from e
    #         logger.debug(
    #             f'Attempting to validate Attr Type {attr} of module \'{module_name}\'.'
    #         )
    #         try:
    #             validate_type(attr_type=getattr(self, attr))
    #         except InvalidAttrTypeException as e:
    #             raise ConfigException(
    #                 f'Invalid Attr Type for {attr} of module \'{module_name}\'. Original validation error: {str(e)}'
    #             ) from e
    #         except InvalidAttrTypeArgException as e:
    #             raise ConfigException(
    #                 f'Invalid Attr Type Arg for {attr} of module \'{module_name}\'. Original validation error: {str(e)}'
    #             ) from e

    logger.debug("Configured module %s", module_name)


async def config_app():

    # [TODO] Re-implement
    # Check API version
    # nawah_level = '.'.join(Config.sys.version.split('.')[0:2])
    # for package, api_level in {
    #     name: config['package'].api_level
    #     for name, config in Config.sys.packages.items()
    # }.items():
    #     if api_level != nawah_level:
    #         raise ConfigException(
    #             f'Nawah framework is on API-level \'{nawah_level}\', but the app package'
    #             f'\'{package}\' requires API-level \'{api_level}\''
    #         )
    # try:
    #     versions = (
    #         (
    #             requests.get(
    #                 'https://raw.githubusercontent.com/masaar/nawah_versions/master/versions.txt'
    #             ).content
    #         )
    #         .decode('utf-8')
    #         .split('\n')
    #     )
    #     version_detected = ''
    #     for version in versions:
    #         if version.startswith(f'{nawah_level}.'):
    #             if version_detected and int(version.split('.')[-1]) < int(
    #                 version_detected.split('.')[-1]
    #             ):
    #                 continue
    #             version_detected = version
    #     if version_detected and version_detected != Config.sys.version:
    #         logger.warning(
    #             f'Your app is using Nawah version \'{Config.sys.version}\' while newer version \'{version_detected}\' of the API-level is available. Please, update'
    #         )
    # except:
    #     logger.warning(
    #         'An error occurred while attempting to check for latest update to Nawah. Please, check for updates on your own'
    #     )

    # Check for jobs
    if Config.jobs:
        # Check jobs schedule validity
        Config.sys.jobs_base = datetime.datetime.utcnow()
        for job_name, job in Config.jobs.items():
            if not croniter.is_valid(job.schedule):
                raise ConfigException(
                    f"Job with schedule '{job_name}' schedule is invalid"
                )

            job.cron_schedule = croniter(job.schedule, Config.sys.jobs_base)
            job.next_time = datetime.datetime.fromtimestamp(
                job.cron_schedule.get_next(), datetime.timezone.utc
            ).isoformat()[:16]

    # Check for presence of user_attrs
    if not Config.user_attrs:
        raise ConfigException("No 'user_attrs' are provided")

    # Check for Env Vars
    attrs_defaults = {
        "data_server": "mongodb://localhost",
        "data_name": "nawah_data",
        "data_ssl": False,
        "emulate_test": False,
    }
    for attr_name in attrs_defaults:
        attr_val = getattr(Config, attr_name)
        if type(attr_val) == str and attr_val.startswith("$__env."):
            logger.debug(f"Detected Env Variable for config attr '{attr_name}'")
            if not os.getenv(attr_val[7:]):
                logger.warning(
                    f"Couldn't read Env Variable for config attr '{attr_name}'. Defaulting to '{attrs_defaults[attr_name]}'"
                )
                setattr(Config, attr_name, attrs_defaults[attr_name])
            else:
                # Set data_ssl to True rather than string Env Variable value
                if attr_name == "ssl":
                    attr_val = True
                else:
                    attr_val = os.getenv(attr_val[7:])
                logger.warning(
                    f"Setting Env Variable for config attr '{attr_name}' to '{attr_val}'"
                )
                setattr(Config, attr_name, attr_val)

    from nawah import data as Data

    # Create default env dict
    anon_user = _compile_anon_user()
    anon_session = _compile_anon_session()
    anon_session["user"] = anon_user
    Config.sys.conn = Data.create_conn()
    Config.sys.session = anon_session
    Config.sys.session.update(
        {
            "conn": {
                "data": Config.sys.conn,
                "REMOTE_ADDR": "127.0.0.1",
                "HTTP_USER_AGENT": "Nawah",
                "client_app": "__sys",
            }
        }
    )

    # Check cache_server Config Attr
    if Config.cache_server:
        logger.debug("Attempting to create connection with 'cache_server'")
        Config.sys.cache = aioredis.Redis(
            host=Config.cache_server,
            db=Config.cache_db,
            username=Config.cache_username,
            password=Config.cache_password,
        ).json()

    # Check error_reporting_server Config Attr
    if Config.error_reporting_server:
        logger.debug("Attempting to create connection with 'error_reporting_server'")
        sentry_sdk.init(  # pylint: disable=abstract-class-instantiated
            Config.error_reporting_server,
            traces_sample_rate=1.0,
            with_locals=True,
            release=f"{Config.sys.name}:{Config.sys.version}",
            environment=Config.env,
        )

    # Check test mode
    if Config.test:
        logger.debug("Test mode detected")
        logger.setLevel(logging.DEBUG)
        __location__ = os.path.realpath(os.path.join("."))
        if not os.path.exists(os.path.join(__location__, "tests")):
            os.makedirs(os.path.join(__location__, "tests"))
        for module_name, module in Config.modules.items():
            module_collection = module.collection
            logger.debug(
                f"Updating collection name '{module_collection}' of module {module_name}"
            )
            module_collection = module.collection = f"test_{module_collection}"

            logger.debug(f"Flushing test collection '{module_collection}'")
            await Data.drop(
                session=Config.sys.session,
                collection_name=module_collection,
            )

    # Checking users collection
    # [TODO] Updated sequence to handle users
    logger.debug("Testing users collection")
    if Config.sys.admin_check:
        user_results = await call(
            "user/read",
            # [TODO] Check behaviour with removed Event.ON
            skip_events=[Event.PERM],
            session=Config.sys.session,
            query=[{"_id": "f00000000000000000000010"}],
            args={
                "skip_sanitise_results": True,
            },
        )
        if not user_results["args"]["count"]:
            logger.debug("ADMIN user not found, creating it")
            # Prepare base ADMIN user doc
            admin_create_doc = {
                "_id": ObjectId("f00000000000000000000010"),
                "name": {Config.locale: "__ADMIN"},
                "groups": [],
                "privileges": {"*": ["*"]},
                "locale": Config.locale,
            }
            # Update ADMIN user doc with admin_doc Config Attr
            admin_create_doc.update(Config.admin_doc)

            for auth_attr in Config.user_attrs:
                admin_create_doc[f"{auth_attr}_hash"] = pbkdf2_sha512.using(
                    rounds=100000
                ).hash(
                    f"{auth_attr}{admin_create_doc[auth_attr]}{Config.admin_password}"
                    f"{Config.anon_token}".encode("utf-8")
                )
            admin_results = await call(
                "user/create",
                skip_events=[Event.PERM],
                session=Config.sys.session,
                doc=admin_create_doc,
            )
            logger.debug("ADMIN user creation results: %s", admin_results)
            if admin_results["status"] != 200:
                raise ConfigException("Config step failed")
        else:
            logger.debug(
                "'ADMIN' doc found. Attempting to check if update is required.."
            )
            admin_doc = user_results["args"]["docs"][0]
            admin_doc_update = {}
            for attr_name, attr_val in Config.admin_doc.items():
                if (
                    attr_name not in admin_doc
                    or not admin_doc[attr_name]
                    or attr_val != admin_doc[attr_name]
                ):
                    if (
                        isinstance(attr_val, dict)
                        and Config.locale in attr_val
                        and isinstance(admin_doc[attr_name], dict)
                        and (
                            (
                                Config.locale in admin_doc[attr_name]
                                and attr_val[Config.locale]
                                == admin_doc[attr_name][Config.locale]
                            )
                            or (Config.locale not in admin_doc[attr_name])
                        )
                    ):
                        continue
                    logger.debug(
                        "Detected change in 'admin_doc.%s' Config Attr", attr_name
                    )
                    admin_doc_update[attr_name] = attr_val
            for auth_attr in Config.user_attrs:
                auth_attr_hash = pbkdf2_sha512.using(rounds=100000).hash(
                    f"{auth_attr}{admin_doc[auth_attr]}{Config.admin_password}"
                    f"{Config.anon_token}".encode("utf-8")
                )
                if (
                    f"{auth_attr}_hash" not in admin_doc
                    or auth_attr_hash != admin_doc[f"{auth_attr}_hash"]
                ):
                    logger.debug(f"Detected change in 'admin_password' Config Attr")
                    admin_doc_update[f"{auth_attr}_hash"] = auth_attr_hash
            if len(admin_doc_update.keys()):
                logger.debug(
                    f"Attempting to update ADMIN user with doc: '{admin_doc_update}'"
                )
                admin_results = await call(
                    "user/update",
                    # [TODO] Check behaviour with removed Event.ON
                    # [TODO] Check behaviour with removed Event.PRE
                    skip_events=[Event.PERM],
                    session=Config.sys.session,
                    query=[{"_id": ObjectId("f00000000000000000000010")}],
                    doc=admin_doc_update,
                )
                logger.debug(f"ADMIN user update results: {admin_results}")
                if admin_results["status"] != 200:
                    raise ConfigException("Config step failed")
            else:
                logger.debug("ADMIN user is up-to-date")
    else:
        logger.warning(
            "Skipping checking 'ADMIN' doc due to 'sys.admin_check' Config Attr"
        )

    Config.sys.docs[ObjectId("f00000000000000000000010")] = SysDoc(module="user")

    # Test if ANON user exists
    user_results = await call(
        "user/read",
        # [TODO] Check behaviour with removed Event.ON
        skip_events=[Event.PERM],
        session=Config.sys.session,
        query=[{"_id": "f00000000000000000000011"}],
        args={
            "skip_sanitise_results": True,
        },
    )
    if not user_results["args"]["count"]:
        logger.debug("ANON user not found, creating it")
        anon_results = await call(
            "user/create",
            # [TODO] Check behaviour with removed Event.ON
            # [TODO] Check behaviour with removed Event.PRE
            skip_events=[Event.PERM],
            session=Config.sys.session,
            doc=_compile_anon_user(),
        )
        logger.debug(f"ANON user creation results: {anon_results}")
        if anon_results["status"] != 200:
            raise ConfigException("Config step failed")
    else:
        logger.debug("ANON user found, checking it")
        anon_doc = _compile_anon_user()
        anon_doc_update = {}
        for attr in Config.user_attrs:
            if attr not in anon_doc or not anon_doc[attr]:
                logger.debug(f"Detected change in 'anon_doc.{attr}' Config Attr")
                anon_doc_update[attr] = generate_attr_val(
                    attr_type=Config.user_attrs[attr].type
                )
        for module in Config.anon_privileges:
            if module not in anon_doc or set(anon_doc[module]) != set(
                Config.anon_privileges[module]
            ):
                logger.debug(f"Detected change in 'anon_privileges' Config Attr")
                anon_doc_update[f"privileges.{module}"] = Config.anon_privileges[module]
        for auth_attr in Config.user_attrs:
            if (
                f"{auth_attr}_hash" not in anon_doc
                or anon_doc[f"{auth_attr}_hash"] != Config.anon_token
            ):
                logger.debug(f"Detected change in 'anon_token' Config Attr")
                anon_doc_update[attr] = Config.anon_token
            anon_doc_update[f"{auth_attr}_hash"] = Config.anon_token
        if len(anon_doc_update.keys()):
            logger.debug(
                f"Attempting to update ANON user with doc: '{anon_doc_update}'"
            )
            anon_results = await call(
                "user/update",
                # [TODO] Check behaviour with removed Event.ON
                # [TODO] Check behaviour with removed Event.PRE
                skip_events=[Event.PERM],
                session=Config.sys.session,
                query=[{"_id": ObjectId("f00000000000000000000011")}],
                doc=anon_doc_update,
            )
            logger.debug(f"ANON user update results: {anon_results}")
            if anon_results["status"] != 200:
                raise ConfigException("Config step failed")
        else:
            logger.debug("ANON user is up-to-date")

    Config.sys.docs[ObjectId("f00000000000000000000011")] = SysDoc(module="user")

    logger.debug("Testing sessions collection")
    # Test if ANON session exists
    session_results = await call(
        "session/read",
        # [TODO] Check behaviour with removed Event.ON
        skip_events=[Event.PERM],
        session=Config.sys.session,
        query=[{"_id": "f00000000000000000000012"}],
    )
    if not session_results["args"]["count"]:
        logger.debug("ANON session not found, creating it")
        anon_results = await call(
            "session/create",
            # [TODO] Check behaviour with removed Event.ON
            # [TODO] Check behaviour with removed Event.PRE
            skip_events=[Event.PERM],
            session=Config.sys.session,
            doc=_compile_anon_session(),
        )
        logger.debug(f"ANON session creation results: {anon_results}")
        if anon_results["status"] != 200:
            raise ConfigException("Config step failed")
    Config.sys.docs[ObjectId("f00000000000000000000012")] = SysDoc(module="session")

    logger.debug("Testing groups collection")
    # Test if DEFAULT group exists
    group_results = await call(
        "group/read",
        # [TODO] Check behaviour with removed Event.ON
        skip_events=[Event.PERM],
        session=Config.sys.session,
        query=[{"_id": "f00000000000000000000013"}],
    )
    if not group_results["args"]["count"]:
        logger.debug("DEFAULT group not found, creating it")
        group_create_doc = {
            "_id": ObjectId("f00000000000000000000013"),
            "user": ObjectId("f00000000000000000000010"),
            "name": {locale: "__DEFAULT" for locale in Config.locales},
            "bio": {locale: "__DEFAULT" for locale in Config.locales},
            "privileges": Config.default_privileges,
        }
        group_results = await call(
            "group/create",
            # [TODO] Check behaviour with removed Event.ON
            # [TODO] Check behaviour with removed Event.PRE
            skip_events=[Event.PERM],
            session=Config.sys.session,
            doc=group_create_doc,
        )
        logger.debug("DEFAULT group creation results: %s", group_results)
        if group_results["status"] != 200:
            raise ConfigException("Config step failed")
    else:
        logger.debug("DEFAULT group found, checking it")
        default_doc = group_results["args"]["docs"][0]
        default_doc_update: MutableMapping[str, Any] = {}
        for module in Config.default_privileges:
            if module not in default_doc["privileges"].keys() or set(
                default_doc["privileges"][module]
            ) != set(Config.default_privileges[module]):
                logger.debug("Detected change in 'default_privileges' Config Attr")
                default_doc_update[f"privileges.{module}"] = Config.default_privileges[
                    module
                ]
        if len(default_doc_update.keys()):
            logger.debug(
                "Attempting to update DEFAULT group with doc: '%s'",
                default_doc_update,
            )
            default_results = await call(
                "group/update",
                # [TODO] Check behaviour with removed Event.ON
                # [TODO] Check behaviour with removed Event.PRE
                skip_events=[Event.PERM],
                session=Config.sys.session,
                query=[{"_id": ObjectId("f00000000000000000000013")}],
                doc=default_doc_update,
            )
            logger.debug(f"DEFAULT group update results: {default_results}")
            if anon_results["status"] != 200:
                raise ConfigException("Config step failed")
        else:
            logger.debug("DEFAULT group is up-to-date")

    Config.sys.docs[ObjectId("f00000000000000000000013")] = SysDoc(module="group")

    # Test app-specific groups
    logger.debug("Testing app-specific groups collection")
    for group in Config.groups:
        group_results = await call(
            "group/read",
            # [TODO] Check behaviour with removed Event.ON
            skip_events=[Event.PERM],
            session=Config.sys.session,
            query=[{"_id": group["_id"]}],
        )
        if not group_results["args"]["count"]:
            logger.debug(
                f'App-specific group with name \'{group["name"]}\' not found, creating it'
            )
            group_results = await call(
                "group/create",
                # [TODO] Check behaviour with removed Event.ON
                # [TODO] Check behaviour with removed Event.PRE
                skip_events=[Event.PERM],
                session=Config.sys.session,
                doc=group,
            )
            logger.debug(
                f'App-specific group with name {group["name"]} creation results: {group_results}'
            )
            if group_results["status"] != 200:
                raise ConfigException("Config step failed")
        else:
            logger.debug(
                f'App-specific group with name \'{group["name"]}\' found, checking it'
            )
            group_doc = group_results["args"]["docs"][0]
            group_doc_update = {}
            if "privileges" in group:
                for module in group["privileges"]:
                    if module not in group_doc["privileges"].keys() or set(
                        group_doc["privileges"][module]
                    ) != set(group["privileges"][module]):
                        logger.debug(
                            f'Detected change in \'privileges\' Doc Arg for group with name \'{group["name"]}\''
                        )
                        group_doc_update[f"privileges.{module}"] = group["privileges"][
                            module
                        ]
            if len(group_doc_update.keys()):
                logger.debug(
                    "Attempting to update group with name '%s' with doc: '%s'",
                    group["name"],
                    group_doc_update,
                )
                group_results = await call(
                    "group/update",
                    # [TODO] Check behaviour with removed Event.ON
                    # [TODO] Check behaviour with removed Event.PRE
                    skip_events=[Event.PERM],
                    session=Config.sys.session,
                    query=[{"_id": group["_id"]}],
                    doc=group_doc_update,
                )
                logger.debug(
                    "Group with name '%s' update results: %s",
                    group["name"],
                    group_results,
                )
                if group_results["status"] != 200:
                    raise ConfigException("Config step failed")
            else:
                logger.debug("Group with name '%s' is up-to-date", group["name"])

        Config.sys.docs[ObjectId(group["_id"])] = SysDoc(module="group")

    # Test app-specific data indexes
    logger.debug("Testing data indexes")
    for index in Config.data_indexes:
        logger.debug("Attempting to create data index: %s", index)
        try:
            Config.sys.conn[Config.data_name][index["collection"]].create_index(
                index["index"]
            )
        except Exception as e:
            logger.error("Failed to create data index: %s, with error: %s", index, e)
            logger.error("Evaluate error and take action manually")

    logger.debug("Creating 'var', 'type', 'user' data indexes for settings collections")
    Config.sys.conn[Config.data_name]["settings"].create_index([("var", 1)])
    Config.sys.conn[Config.data_name]["settings"].create_index([("type", 1)])
    Config.sys.conn[Config.data_name]["settings"].create_index([("user", 1)])
    logger.debug(
        "Creating 'user', 'event', 'subevent' data indexes for analytics collections"
    )
    Config.sys.conn[Config.data_name]["analytics"].create_index([("user", 1)])
    Config.sys.conn[Config.data_name]["analytics"].create_index([("event", 1)])
    Config.sys.conn[Config.data_name]["analytics"].create_index([("subevent", 1)])
    logger.debug("Creating '__deleted' data indexes for all collections")
    for module in Config.modules:
        if Config.modules[module].collection:
            logger.debug(
                "Attempting to create '__deleted' data index for collection: %s",
                Config.modules[module].collection,
            )
            Config.sys.conn[Config.data_name][
                Config.modules[module].collection
            ].create_index([("__deleted", 1)])

    # Test app-specific docs
    logger.debug("Testing docs")
    for doc in Config.docs:
        if not isinstance(doc, SysDoc):
            raise ConfigException("Invalid Config Attr 'docs'")

        doc_results = await call(
            f"{doc.module}/read",
            # [TODO] Check behaviour with removed Event.ON
            # [TODO] Check behaviour with removed Event.PRE
            skip_events=[Event.PERM, Event.ATTRS_QUERY],
            session=Config.sys.session,
            query=[{doc.key: doc.key_value}],  # type: ignore
        )
        if not doc_results["args"]["count"]:
            skip_events = [Event.PERM]
            if doc.skip_args is True:
                skip_events.append(Event.ATTRS_DOC)
            doc.doc = cast("NawahDoc", doc.doc)
            doc_results = await call(
                f"{doc.module}/create",
                skip_events=skip_events,
                session=Config.sys.session,
                doc=doc.doc,
            )
            logger.debug(
                "App-specific doc with %s '%s' of module '%s' creation results: %s",
                doc.key,
                doc.key_value,
                doc.module,
                doc_results,
            )
            if doc_results["status"] != 200:
                raise ConfigException("Config step failed")
        Config.sys.docs[ObjectId(doc_results["args"]["docs"][0]["_id"])] = SysDoc(
            module=doc.module
        )

    # Check for emulate_test mode
    if Config.emulate_test:
        Config.test = True


def _compile_anon_user() -> MutableMapping:
    anon_doc = {
        "_id": ObjectId("f00000000000000000000011"),
        "name": {Config.locale: "__ANON"},
        "groups": [],
        "privileges": Config.anon_privileges,
        "locale": Config.locale,
    }
    for user_attr_name, user_attr in Config.user_attrs.items():
        if user_attr.default:
            anon_doc[user_attr_name] = copy.deepcopy(user_attr.default.value)
        else:
            anon_doc[user_attr_name] = generate_attr_val(attr_type=user_attr.type)
    for auth_attr in Config.auth_attrs:
        anon_doc[f"{auth_attr}_hash"] = Config.anon_token
    if Config.anon_doc:
        anon_doc.update(Config.anon_doc)
    return anon_doc


def _compile_anon_session() -> "NawahSession":
    session_doc: "NawahSession" = {
        "_id": ObjectId("f00000000000000000000012"),
        "user": {"_id": ObjectId("f00000000000000000000011")},
        "groups": [],
        "host_add": "127.0.0.1",
        "user_agent": Config.anon_token,
        "expiry": "1970-01-01T00:00:00",
        "token": Config.anon_token,
        "token_hash": Config.anon_token,
        "create_time": "1970-01-01T00:00:00",
    }
    return session_doc



"""Provides exceptions related to Query"""

from typing import Any, Literal


class InvalidQueryStepTypeException(Exception):
    """Raised by \'validate_query_step\' if type of Query step is not 'dict'"""

    status = 400

    def __init__(self, *, step_type: Any):
        super().__init__(
            f"Query step type should be of type 'dict'. Got '{step_type}' instead"
        )


class InvalidQueryStepLenException(Exception):
    """Raised by \'validate_query_step\' if Query step has more than one item"""

    status = 400

    def __init__(self, *, step_items: tuple[str, ...]):
        super().__init__(f"Query step must have one item. Got '{step_items}' instead")


class InvalidQueryStepAttrTypeException(Exception):
    """Raised by \'validate_query_step\' if type of Query step attr is not 'dict'"""

    status = 400

    def __init__(self, *, step_attr_type: Any):
        super().__init__(
            f"Query step attr value should be of type 'dict'. Got '{step_attr_type}' instead"
        )


class InvalidQueryStepAttrLenException(Exception):
    """Raised by \'validate_query_step\' if Query step attr has more than one item"""

    status = 400

    def __init__(self, *, step_attr_items: tuple[str, ...]):
        super().__init__(
            f"Query step attr can only have one item. Got '{step_attr_items}' instead"
        )


class InvalidQueryOperTypeException(Exception):
    """Raised by \'validate_query_step\' if value of Query step attr is invalid"""

    status = 400

    def __init__(
        self,
        *,
        attr_name: str,
        attr_oper: Literal[
            "$ne",
            "$eq",
            "$gt",
            "$gte",
            "$lt",
            "$lte",
            "$all",
            "$in",
            "$nin",
            "$regex",
        ],
        attr_type: Any,
        attr_val: Any,
    ):
        super().__init__(
            f"Invalid value for Query Arg '{attr_name}' with Query Arg Oper '{attr_oper}'"
            f" expecting type '{attr_type}' but got '{attr_val}'"
        )


class UnknownQueryOperException(Exception):
    """Raised by \'validate_query_step\' if an unknown \'Query Arg\' is detected"""

    status = 400

    def __init__(
        self,
        *,
        attr_name: str,
        attr_oper: Literal[
            "$ne",
            "$eq",
            "$gt",
            "$gte",
            "$lt",
            "$lte",
            "$all",
            "$in",
            "$nin",
            "$regex",
        ],
    ):
        super().__init__(
            f"Unknown Query Arg Oper '{attr_oper}' for Query Arg '{attr_name}'"
        )



"""Provides Exception classes for 'user' Nawah Module"""


class InvalidUserException(Exception):
    """Raised by callable of 'user/read_privileges', 'user/add_group', 'user/delete_group'
    endpoints if user doesn't exist"""

    status = 400

    def __init__(self):
        super().__init__("User is invalid")


class InvalidGroupException(Exception):
    """Raised by callable of 'user/add_group', 'user/delete_group' endpoints if group doesn't
    exist"""

    status = 400

    def __init__(self):
        super().__init__("Group is invalid")


class GroupAddedException(Exception):
    """Raised by callable of 'user/add_group' if user already member in group"""

    status = 400

    def __init__(self):
        super().__init__("User is already a member of the group")


class GroupNotAddedException(Exception):
    """Raised by callable of 'user/delete_group' if user isn't member of group"""

    status = 400

    def __init__(self):
        super().__init__("User is not a member of the group")



import datetime
import logging
import os
import string
from typing import TYPE_CHECKING

from nawah.config import Config
from nawah.enums import AttrType
from nawah.utils._config import setup_app

if TYPE_CHECKING:
    from nawah.classes import App, Attr, Module

logger = logging.getLogger("nawah")


def generate_models_dart(app_config: "App", path: str, filename: str):
    setup_app(app_config)

    # Initialise _api_models Config Attr
    api_models = "// Nawah Models\n"
    # Add interface for DOC, LOCALE, LOCALES, FILE typing
    api_models += "class Doc { String _id; }\n"
    api_models += "class LOCALE { "
    for locale in Config.locales:
        api_models += "String" + ("?" if locale != Config.locale else "") + locale
    api_models += "}\n"
    api_models += (
        "// Locales supported by API '" + "' | '".join(Config.locales) + "';\n"
    )
    api_models += "class LOCALES extends String {}\n"
    api_models += "class ID<T> extends T {}\n"
    api_models += "class FILE { String name; int lastModified; String type; int size; String ref; }\n"
    api_models += (
        "class ATTR { String type; Map args; bool? allow_none; dynamic default; }\n"
    )
    api_models += "class GEO_POINT { String type; List<int> coordinates; }"
    # Iterate over packages in ascending order
    for package in sorted(Config.sys.packages):
        # Add package header
        api_models += f'\n// Package: {package.replace("modules.", "")}\n'
        if not Config.sys.packages[package]:
            api_models += "// No modules\n"
        # Iterate over package modules in ascending order
        for module in sorted(Config.sys.packages[package]["modules"]):
            api_models += _generate_module_model(module=module)
        api_models += "\n"

    models_file = os.path.join(
        os.path.abspath(path),
        filename
        or f'NAWAH_API_MODELS_{datetime.datetime.utcnow().strftime("%d-%b-%Y")}.dart',
    )
    with open(models_file, "w", encoding="UTF-8") as f:
        f.write(api_models)
        logger.info("API models generated and saved to: '%s'", models_file)


def _generate_module_model(*, module: str):
    module_model = ""
    module_class = string.capwords(module.replace("_", " ")).replace(" ", "")
    # Add module header
    module_model += f"// Module: {module_class}\n"

    # Add module interface definition
    module_model += f"class {module_class} extends Doc {{\n"
    # Iterate over module attrs to add attrs types, defaults (if any)
    for attr_name in Config.modules[module].attrs:
        module_model += _generate_attr_model(
            module=module, module_model=module_model, attr_name=attr_name
        )

    # Add closing braces
    module_model += (
        f"\n\t{module_class}({{"
        + ", ".join(
            f"required this.{attr_name}" for attr_name in Config.modules[module].attrs
        )
        + "})\n"
    )
    module_model += (
        "\n\tMap toJson() => {"
        + ", ".join(
            f"'{attr_name}': {attr_name}" for attr_name in Config.modules[module].attrs
        )
        + "}\n"
    )
    module_model += (
        f"\n\t{module_class} fromJson(Map json) => {module_class}("
        + ", ".join(
            f"{attr_name}: json['{attr_name}']"
            for attr_name in Config.modules[module].attrs
        )
        + ")\n"
    )
    module_model += "}\n"

    return module_model


def _generate_attr_model(*, module: str, module_model: str, attr_name: str):
    attr_model = ""
    attr_model += f"\t__TYPE____DEFAULT__ {attr_name};\n"
    for default_attr in Config.modules[module].defaults.keys():
        if default_attr == attr_name or default_attr.startswith(f"{attr_name}."):
            # Attr is in defaults, indicate the same
            attr_model = attr_model.replace("__DEFAULT__", "?")
    # Attempt to replace __DEFAULT__ with empty string if it still exists, effectively no default value
    attr_model = attr_model.replace("__DEFAULT__", "")

    # Add typing
    attr_model = attr_model.replace(
        "__TYPE__",
        _generate_attr_typing(
            module_model=module_model,
            module=Config.modules[module],
            attr_name=attr_name,
            attr_type=Config.modules[module].attrs[attr_name],
        ),
    )

    if Config.modules[module].attrs[attr_name].desc:
        attr_model = (
            f"\t// {Config.modules[module].attrs[attr_name].desc}\n{attr_model}"
        )

    return attr_model


def _generate_attr_typing(
    *, module_model: str, module: "Module", attr_name: str, attr_type: "Attr"
):
    if attr_type.type == AttrType.ANY:
        return "dynamic"

    if attr_type.type == AttrType.BOOL:
        return "bool"

    if attr_type.type == AttrType.DATE:
        return "String"

    if attr_type.type == AttrType.DATETIME:
        return "String"

    if attr_type.type == AttrType.ATTR:
        return "IATTR"

    if attr_type.type == AttrType.KV_DICT:
        return "Map"

    if attr_type.type == AttrType.TYPED_DICT:
        return "Map"

    if attr_type.type == AttrType.EMAIL:
        return "String"

    if attr_type.type == AttrType.FILE:
        return "FILE"

    if attr_type.type == AttrType.FLOAT:
        return "int"

    if attr_type.type == AttrType.GEO_POINT:
        return "GEO_POINT"

    if attr_type.type == AttrType.ID:
        return "ID<String>"

    if attr_type.type == AttrType.INT:
        return "int"

    if attr_type.type == AttrType.IP:
        return "String"

    if attr_type.type == AttrType.LIST:
        return "LIST<dynamic>"

    if attr_type.type == AttrType.LOCALE:
        return "LOCALE"

    if attr_type.type == AttrType.LOCALES:
        return "LOCALES"

    if attr_type.type == AttrType.PHONE:
        return "String"

    if attr_type.type == AttrType.STR:
        return "String"

    if attr_type.type == AttrType.TIME:
        return "String"

    if attr_type.type == AttrType.URI_WEB:
        return "String"

    if attr_type.type == AttrType.LITERAL:
        return "String"

    if attr_type.type == AttrType.UNION:
        return "dynamic"

    if attr_type.type == AttrType.TYPE:
        return "dynamic"



from ._conn import create_conn
from ._create import create
from ._delete import delete
from ._drop import drop
from ._read import read
from ._update import update

__all__ = [
    "create_conn",
    "create",
    "delete",
    "drop",
    "read",
    "update",
]



"""Provides exceptions related to 'call' Utility"""


class InvalidCallEndpointException(Exception):
    """Raised by 'call' Utility if value for arg 'endpoint' is invalid"""

    status = 400

    def __init__(self, *, endpoint):
        super().__init__(
            f"Invliad endpoint format: '{endpoint}'",
        )


class InvalidModuleException(Exception):
    """Raised by 'call' Utility if endpoint points to non-existent Nawah Module"""

    status = 404

    def __init__(self, *, module_name):
        super().__init__(f"Nawah Module '{module_name}' is not defined")


class InvalidFuncException(Exception):
    """Raised by 'call' Utility if endpoint points to non-existent Nawah Function"""

    status = 404

    def __init__(self, *, module_name, func_name):
        super().__init__(
            f"Nawah Module '{module_name}' does not have Function '{func_name}' defined"
        )


class NotPermittedException(Exception):
    """Raised by 'check_permissions' Utility if failed to match Permissions Sets for current
    session user"""

    status = 403

    def __init__(self, *, module_name, func_name):
        super().__init__(f"Not permitted to access '{module_name}'.'{func_name}'")


class InvalidCallArgException(Exception):
    """Raised by 'call' Utility if 'Call Arg' used in call had invalid value"""

    status = 400

    def __init__(self, *, module_name, func_name, arg_name):
        super().__init__(
            f"Invalid value for 'Call Arg' '{arg_name}' of '{module_name}'.'{func_name}'",
        )


class UnknownCallArgException(Exception):
    """Raised by 'call' Utility if 'Call Arg' used in call is not defined in Nawah Function"""

    status = 400

    def __init__(self, *, module_name, func_name, arg_name):
        super().__init__(
            f"Function of '{module_name}'.'{func_name}' does not define '{arg_name}' in 'call_args'"
        )



"""Provides tooling to test Nawah Modules"""

from ._mock import mock_doc, mock_env, mock_results, mock_session, mock_user
from ._setup_test import setup_test

__all__ = [
    "mock_doc",
    "mock_env",
    "mock_results",
    "mock_session",
    "mock_user",
    "setup_test",
]



"""Provides 'delete' Data Function"""

import logging
from typing import TYPE_CHECKING, MutableSequence, Union

from bson import ObjectId

from nawah.config import Config
from nawah.enums import DeleteStrategy
from nawah.exceptions import UnknownDeleteStrategyException

if TYPE_CHECKING:
    from nawah.types import NawahSession, ResultsArgs

logger = logging.getLogger("nawah")


async def delete(
    *,
    session: "NawahSession",
    collection_name: str,
    docs: MutableSequence[Union[str, ObjectId]],
    strategy: DeleteStrategy,
) -> "ResultsArgs":
    """Deletes 'doc' from 'collection' using 'data' connection in 'session'"""

    # Check strategy to cherrypick update, delete calls and system_docs
    if strategy in [DeleteStrategy.SOFT_SKIP_SYS, DeleteStrategy.SOFT_SYS]:
        if strategy == DeleteStrategy.SOFT_SKIP_SYS:
            del_docs = [
                ObjectId(doc) for doc in docs if ObjectId(doc) not in Config.sys.docs
            ]
            if len(del_docs) != len(docs):
                logger.warning(
                    "Skipped soft delete for system docs due to 'DELETE_SOFT_SKIP_SYS' strategy"
                )
        else:
            logger.warning("Detected 'DELETE_SOFT_SYS' strategy for delete call")
            del_docs = [ObjectId(doc) for doc in docs]
        # Perform update call on matching docs
        collection = session["conn"]["data"][Config.data_name][collection_name]
        update_doc = {"$set": {"__deleted": True}}
        # If using Azure Mongo service update docs one by one
        results = await collection.update_many({"_id": {"$in": docs}}, update_doc)
        update_count = results.modified_count

        # Explicitly convert _id value to str to streamline return format across all Data calls
        return {"count": update_count, "docs": [{"_id": str(doc)} for doc in docs]}

    if strategy in [DeleteStrategy.FORCE_SKIP_SYS, DeleteStrategy.FORCE_SYS]:
        if strategy == DeleteStrategy.FORCE_SKIP_SYS:
            del_docs = [
                ObjectId(doc) for doc in docs if ObjectId(doc) not in Config.sys.docs
            ]
            if len(del_docs) != len(docs):
                logger.warning(
                    "Skipped soft delete for system docs due to 'DELETE_FORCE_SKIP_SYS' strategy"
                )
        else:
            logger.warning("Detected 'DELETE_FORCE_SYS' strategy for delete call")
            del_docs = [ObjectId(doc) for doc in docs]
        # Perform delete query on matching docs
        collection = session["conn"]["data"][Config.data_name][collection_name]
        results = await collection.delete_many({"_id": {"$in": del_docs}})
        delete_count = results.deleted_count

        # Explicitly convert _id value to str to streamline return format across all Data calls
        return {"count": delete_count, "docs": [{"_id": str(doc)} for doc in docs]}

    raise UnknownDeleteStrategyException(f"DeleteStrategy '{strategy}' is unknown")



"""Provides 'delete' Base Function callable"""

from typing import TYPE_CHECKING, Optional

import nawah.data as Data
from nawah.config import Config
from nawah.enums import DeleteStrategy, Event

from .exceptions import NoDocDeletedException, UtilityModuleDataCallException

if TYPE_CHECKING:
    from nawah.classes import Query
    from nawah.types import NawahEvents, NawahSession, Results


async def delete(
    *,
    module_name: str,
    skip_events: "NawahEvents",
    session: "NawahSession",
    query: "Query",
    raise_no_success: Optional[bool],
) -> "Results":
    """Deletes doc from module"""

    module = Config.modules[module_name]

    if not module.collection:
        raise UtilityModuleDataCallException(
            module_name=module_name, func_name="delete"
        )

    # [TODO]: confirm all extns are not linked.
    # [DOC] Pick delete strategy based on skip_events
    strategy = DeleteStrategy.SOFT_SKIP_SYS
    if Event.SOFT not in skip_events and Event.SYS_DOCS in skip_events:
        strategy = DeleteStrategy.SOFT_SYS
    elif Event.SOFT in skip_events:
        if Event.SYS_DOCS not in skip_events:
            strategy = DeleteStrategy.FORCE_SKIP_SYS
        else:
            strategy = DeleteStrategy.FORCE_SYS

    docs_results = results = await Data.read(
        session=session,
        collection_name=module.collection,
        attrs=module.attrs,
        query=query,
        skip_process=True,
    )
    results = await Data.delete(
        session=session,
        collection_name=module.collection,
        docs=[doc["_id"] for doc in docs_results["docs"]],
        strategy=strategy,
    )

    if raise_no_success is True and results["count"] == 0:
        raise NoDocDeletedException(module_name=module_name)

    return {"status": 200, "msg": f'Deleted {results["count"]} docs', "args": results}



import copy
import logging
from typing import TYPE_CHECKING, Any, MutableMapping, MutableSequence, Union

from bson import ObjectId

from nawah.config import Config

if TYPE_CHECKING:
    from nawah.types import NawahDoc, NawahSession, ResultsArgs

logger = logging.getLogger("nawah")


async def update(
    *,
    session: "NawahSession",
    collection_name: str,
    docs: MutableSequence[Union[str, ObjectId]],
    doc: "NawahDoc",
) -> "ResultsArgs":
    # Recreate docs list by converting all docs items to ObjectId
    docs = [ObjectId(doc) for doc in docs]
    # Perform update query on matching docs
    collection = session["conn"]["data"][Config.data_name][collection_name]
    results = None
    doc = copy.deepcopy(doc)

    # [TODO] Abstract $set pipeline with colon support for all stages

    # Prepare empty update pipeline
    update_pipeline = []

    # Iterate over attrs in doc to set update stage in pipeline
    for attr in doc.keys():
        # Prepare stage pipeline
        update_pipeline_stage_root: MutableMapping[str, Any] = {"$set": {}}
        update_pipeline_stage_current = update_pipeline_stage_root["$set"]
        attr_path_part = attr
        attr_path_current = []

        if ":" in attr:
            attr_path = attr.split(".")
            for i, attr_path_part in enumerate(attr_path):
                if i == 0:
                    # First item has to have $
                    attr_path_current.append("$" + attr_path_part.split(":")[0])
                else:
                    attr_path_current.append(attr_path_part.split(":")[0])

                if ":" not in attr_path_part:
                    part_pipeline: MutableMapping[str, Any] = {
                        "$arrayToObject": {
                            "$concatArrays": [
                                {
                                    "$objectToArray": ".".join(attr_path_current[:-1]),
                                },
                                [
                                    {
                                        "k": attr_path_part,
                                        "v": None,
                                    }
                                ],
                            ]
                        }
                    }
                    if "v" in update_pipeline_stage_current:
                        update_pipeline_stage_current["v"] = part_pipeline
                    elif "then" in update_pipeline_stage_current:
                        update_pipeline_stage_current["then"] = part_pipeline
                    else:
                        update_pipeline_stage_current[attr_path_part] = part_pipeline

                    update_pipeline_stage_current = part_pipeline["$arrayToObject"][
                        "$concatArrays"
                    ][1][0]
                else:
                    part_pipeline = {
                        "$map": {
                            "input": ".".join(attr_path_current),
                            "as": f"this_{i}",
                            "in": {
                                "$cond": {
                                    "if": {
                                        "$eq": [
                                            {
                                                "$indexOfArray": [
                                                    ".".join(attr_path_current),
                                                    f"$$this_{i}",
                                                ]
                                            },
                                            int(attr_path_part.split(":")[1]),
                                        ]
                                    },
                                    "then": None,
                                    "else": f"$$this_{i}",
                                }
                            },
                        }
                    }

                    if i != 0:
                        # For all subsequent array objects, wrap in object-to-array-to-object pipeline
                        part_pipeline = {
                            "$arrayToObject": {
                                "$concatArrays": [
                                    {
                                        "$objectToArray": ".".join(
                                            attr_path_current[:-1]
                                        ),
                                    },
                                    [
                                        {
                                            "k": attr_path_part.split(":")[0],
                                            "v": part_pipeline,
                                        }
                                    ],
                                ]
                            }
                        }

                    if "v" in update_pipeline_stage_current:
                        update_pipeline_stage_current["v"] = part_pipeline
                    elif "then" in update_pipeline_stage_current:
                        update_pipeline_stage_current["then"] = part_pipeline
                    else:
                        update_pipeline_stage_current[
                            attr_path_part.split(":")[0]
                        ] = part_pipeline

                    if i == 0:
                        update_pipeline_stage_current = part_pipeline["$map"]["in"][
                            "$cond"
                        ]
                    else:
                        update_pipeline_stage_current = part_pipeline["$arrayToObject"][
                            "$concatArrays"
                        ][1][0]["v"]["$map"]["in"]["$cond"]

                    attr_path_current = [f"$$this_{i}"]

        # Check for $add Doc Oper
        if isinstance(doc[attr], dict) and "$add" in doc[attr]:
            add_field = (
                f'${doc[attr]["$field"]}'
                if "$field" in doc[attr].keys() and doc[attr]["$field"]
                else f'${".".join(attr_path_current + [attr])}'
            )

            part_pipeline = {
                "$add": [
                    {
                        "$cond": {
                            "if": {"$not": [add_field]},
                            "then": 0,
                            "else": add_field,
                        }
                    },
                    doc[attr]["$add"],
                ]
            }

            # Add part_pipeline to update_pipeline_stage_current
            if "v" in update_pipeline_stage_current.keys():
                update_pipeline_stage_current["v"] = part_pipeline
            elif "then" in update_pipeline_stage_current.keys():
                update_pipeline_stage_current["then"] = part_pipeline
            else:
                update_pipeline_stage_current[attr_path_part] = part_pipeline

        # Check for $add Doc Oper
        elif isinstance(doc[attr], dict) and "$multiply" in doc[attr].keys():
            multiply_field = (
                f'${doc[attr]["$field"]}'
                if "$field" in doc[attr].keys() and doc[attr]["$field"]
                else f'${".".join(attr_path_current + [attr])}'
            )

            part_pipeline = {
                "$multiply": [
                    {
                        "$cond": {
                            "if": {"$not": [multiply_field]},
                            "then": 0,
                            "else": multiply_field,
                        }
                    },
                    doc[attr]["$multiply"],
                ]
            }

            # Add part_pipeline to update_pipeline_stage_current
            if "v" in update_pipeline_stage_current.keys():
                update_pipeline_stage_current["v"] = part_pipeline
            elif "then" in update_pipeline_stage_current.keys():
                update_pipeline_stage_current["then"] = part_pipeline
            else:
                update_pipeline_stage_current[attr_path_part] = part_pipeline

        # Check for $append Doc Oper
        elif isinstance(doc[attr], dict) and "$append" in doc[attr].keys():
            if "$unique" not in doc[attr].keys() or doc[attr]["$unique"] is False:
                part_pipeline = {"$concatArrays": [f"${attr}", [doc[attr]["$append"]]]}
            else:
                part_pipeline = {
                    "$concatArrays": [
                        f"${attr}",
                        {
                            "$cond": {
                                "if": {"$in": [doc[attr]["$append"], f"${attr}"]},
                                "then": [],
                                "else": [doc[attr]["$append"]],
                            }
                        },
                    ]
                }

            # Add part_pipeline to update_pipeline_stage_current
            if "v" in update_pipeline_stage_current.keys():
                update_pipeline_stage_current["v"] = part_pipeline
            elif "then" in update_pipeline_stage_current.keys():
                update_pipeline_stage_current["then"] = part_pipeline
            else:
                update_pipeline_stage_current[attr_path_part] = part_pipeline

        # Check for $set_index Doc Oper
        elif isinstance(doc[attr], dict) and "$set_index" in doc[attr]:
            part_pipeline = {
                "$reduce": {
                    "input": f"${attr}",
                    "initialValue": [],
                    "in": {
                        "$concatArrays": [
                            "$$value",
                            {
                                "$cond": {
                                    "if": {
                                        "$eq": [
                                            ["$$this"],
                                            [
                                                {
                                                    "$arrayElemAt": [
                                                        f"${attr}",
                                                        doc[attr]["$index"],
                                                    ]
                                                }
                                            ],
                                        ]
                                    },
                                    "then": [doc[attr]["$set_index"]],
                                    "else": ["$$this"],
                                }
                            },
                        ]
                    },
                }
            }

            # Add part_pipeline to update_pipeline_stage_current
            if "v" in update_pipeline_stage_current.keys():
                update_pipeline_stage_current["v"] = part_pipeline
            elif "then" in update_pipeline_stage_current.keys():
                update_pipeline_stage_current["then"] = part_pipeline
            else:
                update_pipeline_stage_current[attr_path_part] = part_pipeline

        # Check for $del_val Doc Oper
        elif isinstance(doc[attr], dict) and "$del_val" in doc[attr]:
            part_pipeline = {
                "$reduce": {
                    "input": f"${attr}",
                    "initialValue": [],
                    "in": {
                        "$concatArrays": [
                            "$$value",
                            {
                                "$cond": {
                                    "if": {
                                        "$eq": [
                                            ["$$this"],
                                            doc[attr]["$del_val"],
                                        ]
                                    },
                                    "then": [],
                                    "else": ["$$this"],
                                }
                            },
                        ]
                    },
                }
            }

            # Add part_pipeline to update_pipeline_stage_current
            if "v" in update_pipeline_stage_current.keys():
                update_pipeline_stage_current["v"] = part_pipeline
            elif "then" in update_pipeline_stage_current.keys():
                update_pipeline_stage_current["then"] = part_pipeline
            else:
                update_pipeline_stage_current[attr_path_part] = part_pipeline

        # Check for $del_index Doc Oper
        elif isinstance(doc[attr], dict) and "$del_index" in doc[attr]:
            part_pipeline = {
                "$arrayToObject": {
                    "$reduce": {
                        "input": {
                            "$objectToArray": f"${attr}",
                        },
                        "initialValue": [],
                        "in": {
                            "$concatArrays": [
                                "$$value",
                                {
                                    "$cond": {
                                        "if": {
                                            "$eq": [
                                                "$$this.k",
                                                doc[attr]["$del_index"],
                                            ],
                                        },
                                        "then": [],
                                        "else": ["$$this"],
                                    }
                                },
                            ]
                        },
                    }
                }
            }

            # Add part_pipeline to update_pipeline_stage_current
            if "v" in update_pipeline_stage_current.keys():
                update_pipeline_stage_current["v"] = part_pipeline
            elif "then" in update_pipeline_stage_current.keys():
                update_pipeline_stage_current["then"] = part_pipeline
            else:
                update_pipeline_stage_current[attr_path_part] = part_pipeline

        else:
            # Add part_pipeline to update_pipeline_stage_current
            if "v" in update_pipeline_stage_current.keys():
                update_pipeline_stage_current["v"] = {"$literal": doc[attr]}
            elif "then" in update_pipeline_stage_current.keys():
                update_pipeline_stage_current["then"] = {"$literal": doc[attr]}
            else:
                update_pipeline_stage_current[attr_path_part] = {"$literal": doc[attr]}

        # Add stage to pipeline
        update_pipeline.append(update_pipeline_stage_root)

    logger.debug("Final update query: %s", {"_id": {"$in": docs}})
    logger.debug("Final update pipeline: %s", update_pipeline)

    results = await collection.update_many({"_id": {"$in": docs}}, update_pipeline)
    update_count = results.modified_count

    # Explicitly convert _id value to str to streamline return format across all Data calls
    return {"count": update_count, "docs": [{"_id": str(doc)} for doc in docs]}



"""Nawah is rapid app development framework"""

import logging
import os

logger = logging.getLogger("nawah")
handler = logging.StreamHandler()
formatter = logging.Formatter("%(asctime)s [%(levelname)s]  %(message)s")
handler.setFormatter(formatter)
logger.addHandler(handler)

logger.setLevel(logging.INFO)

__version__ = "0.0.0"
with open(
    os.path.join(os.path.dirname(__file__), "version.txt"), encoding="UTF-8"
) as f:
    __version__ = f.read().strip()



"""Provides exceptions related to validation"""


class MissingAttrException(Exception):
    """Raised by 'validate_doc' Utility if required 'attr' is missing from 'doc'"""

    status = 400

    def __init__(self, *, attr_name):
        super().__init__(f"Missing attr '{attr_name}'")


class InvalidAttrException(Exception):
    """Raised by 'validate_attr' Utility if 'attr' has invalid value"""

    status = 400

    def __init__(self, *, attr_name, attr_type, val_type):
        super().__init__(
            f"Invalid attr '{attr_name}' of type '{val_type}' with required type '{attr_type}'"
        )



"""Provides 'delete_lock' Base Function callable"""

import logging
from typing import TYPE_CHECKING

import nawah.data as Data
from nawah.config import Config
from nawah.enums import DeleteStrategy

from .exceptions import FailedDeleteLockException

if TYPE_CHECKING:
    from nawah.classes import Query
    from nawah.types import NawahSession, Results

logger = logging.getLogger("nawah")


async def delete_lock(
    *, module_name: str, session: "NawahSession", query: "Query"
) -> "Results":
    """Deletes locks for a module matching query \'query\'. If not, raises MethodException."""

    lock_id = query["_id:$eq"][0]

    module = Config.modules[module_name]

    docs_results = results = await Data.read(
        session=session,
        collection_name=f"{module.collection}__lock",
        attrs={},
        query=query,
        skip_process=True,
    )
    results = await Data.delete(
        session=session,
        collection_name=f"{module.collection}__lock",
        docs=[doc["_id"] for doc in docs_results["docs"]],
        strategy=DeleteStrategy.FORCE_SYS,
    )

    if results["count"] != 1:
        logger.error(
            "Failed to delete lock '%s' for '%s'. DELETE MANUALLY NOW",
            lock_id,
            module_name,
        )
        raise FailedDeleteLockException(module_name=module_name, lock_id=lock_id)

    return {
        "status": 200,
        "msg": f'Deleted {results["count"]} docs',
        "args": results,
    }



import aiohttp.web
from multidict import MultiDict

from nawah.classes import app_encoder
from nawah.config import Config


async def _not_found_handler(_):
    headers = MultiDict(
        [
            ("Server", "Nawah"),
            ("Powered-By", "Nawah, https://nawah.masaar.com"),
            ("Access-Control-Allow-Origin", "*"),
            ("Access-Control-Allow-Methods", "GET,POST"),
            ("Access-Control-Allow-Headers", "Content-Type"),
            ("Access-Control-Expose-Headers", "Content-Disposition"),
        ]
    )
    return aiohttp.web.Response(
        status=404,
        headers=headers,
        body=app_encoder.encode({"status": 404, "msg": "404 NOT FOUND"}),
    )


async def _not_allowed_handler(_):
    headers = MultiDict(
        [
            ("Server", "Nawah"),
            ("Powered-By", "Nawah, https://nawah.masaar.com"),
            ("Access-Control-Allow-Origin", "*"),
            ("Access-Control-Allow-Methods", "*"),
            ("Access-Control-Allow-Headers", "Content-Type"),
            ("Access-Control-Expose-Headers", "Content-Disposition"),
        ]
    )
    return aiohttp.web.Response(
        status=405,
        headers=headers,
        body=app_encoder.encode({"status": 405, "msg": "405 NOT ALLOWED"}),
    )


async def _root_handler(_):
    headers = MultiDict(
        [
            ("Server", "Nawah"),
            ("Powered-By", "Nawah, https://nawah.masaar.com"),
            ("Access-Control-Allow-Origin", "*"),
            ("Access-Control-Allow-Methods", "GET"),
            ("Access-Control-Allow-Headers", "Content-Type"),
            ("Access-Control-Expose-Headers", "Content-Disposition"),
        ]
    )
    return aiohttp.web.Response(
        status=200,
        headers=headers,
        body=app_encoder.encode(
            {
                "status": 200,
                "msg": f"Welcome to {Config.sys.name}!",
                "args": {"version": Config.sys.version, "powered_by": "Nawah"},
            }
        ),
    )



from dataclasses import dataclass
from typing import Any, MutableMapping, Optional

import requests


class UrlNotSetException(Exception):
    pass


class UserAgentNotSetException(Exception):
    pass


class ForwardedNotSetException(Exception):
    pass


class EventRequestException(Exception):
    pass


# lyh1UwLGd5XVgsuTe4zDNVwXshlifHAEfP_dy6KEIA9W2Ur7sBVe4PSHMmTEMmab


@dataclass(kw_only=True)
class Plausible:

    host: str
    app_domain: str
    ssl: bool = True
    default_url: Optional[str] = None
    default_user_agent: Optional[str] = None
    default_forwarded: Optional[str] = None

    def register_event(
        self,
        *,
        name="pageview",
        url: str = None,
        user_agent: str = None,
        forwarded: str = None,
        screen_width: str = None,
        props: MutableMapping[str, Any] = None,
    ):
        if not (url or self.default_url):
            raise UrlNotSetException()

        if not (user_agent or self.default_user_agent):
            raise UserAgentNotSetException()

        if not (forwarded or self.default_forwarded):
            raise ForwardedNotSetException()

        api_uri = f'http{"s" if self.ssl else ""}://{self.host}/api/event'
        r = requests.post(
            url=api_uri,
            headers={
                "User-Agent": user_agent or self.default_user_agent,
                "X-Forwarded-For": forwarded or self.default_forwarded,
            },
            json={
                "name": name,
                "url": url,
                "domain": self.app_domain,
                "screen_width": screen_width,
                "props": props or {},
            },
        )

        if r.status_code != 202:
            raise EventRequestException(r.text)
